# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

from inspect import isclass

import torch
import torch.nn as nn

from pyro.distributions.util import broadcast_shape


class Exp(nn.Module):
    """
    a custom module for exponentiation of tensors
    """

    def __init__(self):
        super().__init__()

    def forward(self, val):
        return torch.exp(val)


class ConcatModule(nn.Module):
    """
    a custom module for concatenation of tensors
    """

    def __init__(self, allow_broadcast=False):
        self.allow_broadcast = allow_broadcast
        super().__init__()

    def forward(self, *input_args):
        # we have a single object
        if len(input_args) == 1:
            # regardless of type,
            # we don't care about single objects
            # we just index into the object
            input_args = input_args[0]

        # don't concat things that are just single objects
        if torch.is_tensor(input_args):
            return input_args
        else:
            if self.allow_broadcast:
                shape = broadcast_shape(*[s.shape[:-1] for s in input_args]) + (-1,)
                input_args = [s.expand(shape) for s in input_args]
            return torch.cat(input_args, dim=-1)


class ListOutModule(nn.ModuleList):
    """
    a custom module for outputting a list of tensors from a list of nn modules
    """

    def __init__(self, modules):
        super().__init__(modules)

    def forward(self, *args, **kwargs):
        # loop over modules in self, apply same args
        return [mm.forward(*args, **kwargs) for mm in self]


def call_nn_op(op):
    """
    a helper function that adds appropriate parameters when calling
    an nn module representing an operation like Softmax

    :param op: the nn.Module operation to instantiate
    :return: instantiation of the op module with appropriate parameters
    """
    if op in [nn.Softmax, nn.LogSoftmax]:
        return op(dim=1)
    else:
        return op()


class MLP(nn.Module):
    def __init__(
        self,
        mlp_sizes,
        activation=nn.ReLU,
        output_activation=None,
        post_layer_fct=lambda layer_ix, total_layers, layer: None,
        post_act_fct=lambda layer_ix, total_layers, layer: None,
        allow_broadcast=False,
        use_cuda=False,
    ):
        # init the module object
        super().__init__()

        assert len(mlp_sizes) >= 2, "Must have input and output layer sizes defined"

        # get our inputs, outputs, and hidden
        input_size, hidden_sizes, output_size = (
            mlp_sizes[0],
            mlp_sizes[1:-1],
            mlp_sizes[-1],
        )

        # assume int or list
        assert isinstance(
            input_size, (int, list, tuple)
        ), "input_size must be int, list, tuple"

        # everything in MLP will be concatted if it's multiple arguments
        last_layer_size = input_size if type(input_size) == int else sum(input_size)

        # everything sent in will be concatted together by default
        all_modules = [ConcatModule(allow_broadcast)]

        # loop over l
        for layer_ix, layer_size in enumerate(hidden_sizes):
            assert type(layer_size) == int, "Hidden layer sizes must be ints"

            # get our nn layer module (in this case nn.Linear by default)
            cur_linear_layer = nn.Linear(last_layer_size, layer_size)

            # for numerical stability -- initialize the layer properly
            cur_linear_layer.weight.data.normal_(0, 0.001)
            cur_linear_layer.bias.data.normal_(0, 0.001)

            # use GPUs to share data during training (if available)
            if use_cuda:
                cur_linear_layer = nn.DataParallel(cur_linear_layer)

            # add our linear layer
            all_modules.append(cur_linear_layer)

            # handle post_linear
            post_linear = post_layer_fct(
                layer_ix + 1, len(hidden_sizes), all_modules[-1]
            )

            # if we send something back, add it to sequential
            # here we could return a batch norm for example
            if post_linear is not None:
                all_modules.append(post_linear)

            # handle activation (assumed no params -- deal with that later)
            all_modules.append(activation())

            # now handle after activation
            post_activation = post_act_fct(
                layer_ix + 1, len(hidden_sizes), all_modules[-1]
            )

            # handle post_activation if not null
            # could add batch norm for example
            if post_activation is not None:
                all_modules.append(post_activation)

            # save the layer size we just created
            last_layer_size = layer_size

        # now we have all of our hidden layers
        # we handle outputs
        assert isinstance(
            output_size, (int, list, tuple)
        ), "output_size must be int, list, tuple"

        if type(output_size) == int:
            all_modules.append(nn.Linear(last_layer_size, output_size))
            if output_activation is not None:
                all_modules.append(
                    call_nn_op(output_activation)
                    if isclass(output_activation)
                    else output_activation
                )
        else:

            # we're going to have a bunch of separate layers we can spit out (a tuple of outputs)
            out_layers = []

            # multiple outputs? handle separately
            for out_ix, out_size in enumerate(output_size):

                # for a single output object, we create a linear layer and some weights
                split_layer = []

                # we have an activation function
                split_layer.append(nn.Linear(last_layer_size, out_size))

                # then we get our output activation (either we repeat all or we index into a same sized array)
                act_out_fct = (
                    output_activation
                    if not isinstance(output_activation, (list, tuple))
                    else output_activation[out_ix]
                )

                if act_out_fct:
                    # we check if it's a class. if so, instantiate the object
                    # otherwise, use the object directly (e.g. pre-instaniated)
                    split_layer.append(
                        call_nn_op(act_out_fct) if isclass(act_out_fct) else act_out_fct
                    )

                # our outputs is just a sequential of the two
                out_layers.append(nn.Sequential(*split_layer))

            all_modules.append(ListOutModule(out_layers))

        # now we have all of our modules, we're ready to build our sequential!
        # process mlps in order, pretty standard here
        self.sequential_mlp = nn.Sequential(*all_modules)

    # pass through our sequential for the output!
    def forward(self, *args, **kwargs):
        return self.sequential_mlp.forward(*args, **kwargs)



# Importing specific functions from modules
from .utils import tensor_to_numpy, move_to_device, convert_to_tensor
from .utils import CustomDataset, CustomDataset2, CustomDataset3
from .utils import CustomMultiOmicsDataset, CustomMultiOmicsDataset2
from .utils import pretty_print, Colors
from .utils import find_partitions_greedy

from .queue import PriorityQueue

from .custom_mlp import MLP, Exp

# Importing modules
#from . import utils
#from . import custom_mlp

#__all__ = ['tensor_to_numpy', 'move_to_device', 'convert_to_tensor',
#           'CustomDataset', 'CustomDataset2', 'CustomDataset3',
#           'MLP','Exp',
#           'custom_mlp','utils']


import pyro
import pyro.distributions as dist
from pyro.optim import ExponentialLR
from pyro.infer import SVI, JitTraceEnum_ELBO, TraceEnum_ELBO, config_enumerate

import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from torch.distributions.utils import logits_to_probs, probs_to_logits, clamp_probs
from torch.distributions import constraints
from torch.distributions.transforms import SoftmaxTransform

from .utils.custom_mlp import MLP, Exp
from .utils.utils import CustomDataset, CustomDataset3, CustomDataset4, tensor_to_numpy, convert_to_tensor


import os
import argparse
import random
import numpy as np
import datatable as dt
from tqdm import tqdm
from scipy import sparse

from typing import Literal

import warnings
warnings.filterwarnings("ignore")

import dill as pickle
import gzip
from packaging.version import Version
torch_version = torch.__version__


def set_random_seed(seed):
    # Set seed for PyTorch
    torch.manual_seed(seed)
    
    # If using CUDA, set the seed for CUDA
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)  # For multi-GPU setups.
    
    # Set seed for NumPy
    np.random.seed(seed)
    
    # Set seed for Python's random module
    random.seed(seed)

    # Set seed for Pyro
    pyro.set_rng_seed(seed)

class SURE(nn.Module):
    """SUccinct REpresentation of single-omics cells

    Parameters
    ----------
    inpute_size
        Number of features (e.g., genes, peaks, proteins, etc.) per cell.
    undesired_size
        Number of undesired factors. It would be used to adjust for undesired variations like batch effect.
    codebook_size
        Number of metacells.
    latent_dim
        Dimensionality of latent states and metacells. 
    hidden_layers
        A list give the numbers of neurons for each hidden layer.
    loss_func
        The likelihood model for single-cell data generation. 
        
        One of the following: 
        * ``'negbinomial'`` -  negative binomial distribution (default)
        * ``'poisson'`` - poisson distribution
        * ``'multinomial'`` - multinomial distribution
    user_dirichlet
        A boolean option. If toggled on, SURE characterizes single-cell data using a hierarchical model, such as
        dirichlet-negative binomial.
    latent_dist
        The distribution model for latent states. 
        
        One of the following:
        * ``'normal'`` - normal distribution
        * ``'laplacian'`` - Laplacian distribution
        * ``'studentt'`` - Student-t distribution. 
    use_cuda
        A boolean option for switching on cuda device. 

    Examples
    --------
    >>>
    >>>
    >>>

    """
    def __init__(self,
                 input_size: int,
                 codebook_size: int = 200,
                 cell_factor_size: int = 0,
                 supervised_mode: bool = False,
                 z_dim: int = 10,
                 z_dist: Literal['normal','studentt','laplacian','cauchy','gumbel'] = 'normal',
                 loss_func: Literal['negbinomial','poisson','multinomial'] = 'negbinomial',
                 inverse_dispersion: float = 10.0,
                 use_zeroinflate: bool = False,
                 hidden_layers: list = [500],
                 hidden_layer_activation: Literal['relu','softplus','leakyrelu','linear'] = 'relu',
                 nn_dropout: float = 0.1,
                 post_layer_fct: list = ['layernorm'],
                 post_act_fct: list = None,
                 config_enum: str = 'parallel',
                 use_cuda: bool = False,
                 seed: int = 42,
                 dtype = torch.float32, # type: ignore
                 ):
        super().__init__()

        self.input_size = input_size
        self.cell_factor_size = cell_factor_size
        self.inverse_dispersion = inverse_dispersion
        self.latent_dim = z_dim
        self.hidden_layers = hidden_layers
        self.decoder_hidden_layers = hidden_layers[::-1]
        self.allow_broadcast = config_enum == 'parallel'
        self.use_cuda = use_cuda
        self.loss_func = loss_func
        self.options = None
        self.code_size=codebook_size
        self.supervised_mode=supervised_mode
        self.latent_dist = z_dist
        self.dtype = dtype
        self.use_zeroinflate=use_zeroinflate
        self.nn_dropout = nn_dropout
        self.post_layer_fct = post_layer_fct
        self.post_act_fct = post_act_fct
        self.hidden_layer_activation = hidden_layer_activation
        
        self.codebook_weights = None

        set_random_seed(seed)
        self.setup_networks()

    def setup_networks(self):
        latent_dim = self.latent_dim
        hidden_sizes = self.hidden_layers

        nn_layer_norm, nn_batch_norm, nn_layer_dropout = False, False, False
        na_layer_norm, na_batch_norm, na_layer_dropout = False, False, False

        if self.post_layer_fct is not None:
            nn_layer_norm=True if ('layernorm' in self.post_layer_fct) or ('layer_norm' in self.post_layer_fct) else False
            nn_batch_norm=True if ('batchnorm' in self.post_layer_fct) or ('batch_norm' in self.post_layer_fct) else False
            nn_layer_dropout=True if 'dropout' in self.post_layer_fct else False

        if self.post_act_fct is not None:
            na_layer_norm=True if ('layernorm' in self.post_act_fct) or ('layer_norm' in self.post_act_fct) else False
            na_batch_norm=True if ('batchnorm' in self.post_act_fct) or ('batch_norm' in self.post_act_fct) else False
            na_layer_dropout=True if 'dropout' in self.post_act_fct else False

        if nn_layer_norm and nn_batch_norm and nn_layer_dropout:
            post_layer_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.Dropout(self.nn_dropout),nn.BatchNorm1d(layer.module.out_features), nn.LayerNorm(layer.module.out_features))
        elif nn_layer_norm and nn_layer_dropout:
            post_layer_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.Dropout(self.nn_dropout), nn.LayerNorm(layer.module.out_features))
        elif nn_batch_norm and nn_layer_dropout:
            post_layer_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.Dropout(self.nn_dropout), nn.BatchNorm1d(layer.module.out_features))
        elif nn_layer_norm and nn_batch_norm:
            post_layer_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.BatchNorm1d(layer.module.out_features), nn.LayerNorm(layer.module.out_features))
        elif nn_layer_norm:
            post_layer_fct = lambda layer_ix, total_layers, layer: nn.LayerNorm(layer.module.out_features)
        elif nn_batch_norm:
            post_layer_fct = lambda layer_ix, total_layers, layer:nn.BatchNorm1d(layer.module.out_features)
        elif nn_layer_dropout:
            post_layer_fct = lambda layer_ix, total_layers, layer: nn.Dropout(self.nn_dropout)
        else:
            post_layer_fct = lambda layer_ix, total_layers, layer: None

        if na_layer_norm and na_batch_norm and na_layer_dropout:
            post_act_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.Dropout(self.nn_dropout),nn.BatchNorm1d(layer.module.out_features), nn.LayerNorm(layer.module.out_features))
        elif na_layer_norm and na_layer_dropout:
            post_act_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.Dropout(self.nn_dropout), nn.LayerNorm(layer.module.out_features))
        elif na_batch_norm and na_layer_dropout:
            post_act_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.Dropout(self.nn_dropout), nn.BatchNorm1d(layer.module.out_features))
        elif na_layer_norm and na_batch_norm:
            post_act_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.BatchNorm1d(layer.module.out_features), nn.LayerNorm(layer.module.out_features))
        elif na_layer_norm:
            post_act_fct = lambda layer_ix, total_layers, layer: nn.LayerNorm(layer.module.out_features)
        elif na_batch_norm:
            post_act_fct = lambda layer_ix, total_layers, layer:nn.BatchNorm1d(layer.module.out_features)
        elif na_layer_dropout:
            post_act_fct = lambda layer_ix, total_layers, layer: nn.Dropout(self.nn_dropout)
        else:
            post_act_fct = lambda layer_ix, total_layers, layer: None

        if self.hidden_layer_activation == 'relu':
            activate_fct = nn.ReLU
        elif self.hidden_layer_activation == 'softplus':
            activate_fct = nn.Softplus
        elif self.hidden_layer_activation == 'leakyrelu':
            activate_fct = nn.LeakyReLU
        elif self.hidden_layer_activation == 'linear':
            activate_fct = nn.Identity

        if self.supervised_mode:
            self.encoder_n = MLP(
                [self.input_size] + hidden_sizes + [self.code_size],
                activation=activate_fct,
                output_activation=None,
                post_layer_fct=post_layer_fct,
                post_act_fct=post_act_fct,
                allow_broadcast=self.allow_broadcast,
                use_cuda=self.use_cuda,
            )
        else:
            self.encoder_n = MLP(
                [self.latent_dim] + hidden_sizes + [self.code_size],
                activation=activate_fct,
                output_activation=None,
                post_layer_fct=post_layer_fct,
                post_act_fct=post_act_fct,
                allow_broadcast=self.allow_broadcast,
                use_cuda=self.use_cuda,
            )

        self.encoder_zn = MLP(
            [self.input_size] + hidden_sizes + [[latent_dim, latent_dim]],
            activation=activate_fct,
            output_activation=[None, Exp],
            post_layer_fct=post_layer_fct,
            post_act_fct=post_act_fct,
            allow_broadcast=self.allow_broadcast,
            use_cuda=self.use_cuda,
        )

        if self.cell_factor_size>0:
            self.cell_factor_effect = MLP(
                    [self.cell_factor_size] + self.decoder_hidden_layers + [self.latent_dim],
                    activation=activate_fct,
                    output_activation=None,
                    post_layer_fct=post_layer_fct,
                    post_act_fct=post_act_fct,
                    allow_broadcast=self.allow_broadcast,
                    use_cuda=self.use_cuda,
                )
            
        self.decoder_concentrate = MLP(
                [self.latent_dim] + self.decoder_hidden_layers + [self.input_size],
                activation=activate_fct,
                output_activation=None,
                post_layer_fct=post_layer_fct,
                post_act_fct=post_act_fct,
                allow_broadcast=self.allow_broadcast,
                use_cuda=self.use_cuda,
            )

        if self.latent_dist == 'studentt':
            self.codebook = MLP(
                [self.code_size] + hidden_sizes + [[latent_dim,latent_dim]],
                activation=activate_fct,
                output_activation=[Exp,None],
                post_layer_fct=post_layer_fct,
                post_act_fct=post_act_fct,
                allow_broadcast=self.allow_broadcast,
                use_cuda=self.use_cuda,
            )
        else:
            self.codebook = MLP(
                [self.code_size] + hidden_sizes + [latent_dim],
                activation=activate_fct,
                output_activation=None,
                post_layer_fct=post_layer_fct,
                post_act_fct=post_act_fct,
                allow_broadcast=self.allow_broadcast,
                use_cuda=self.use_cuda,
            )

        if self.use_cuda:
            self.cuda()

    def get_device(self):
        return next(self.parameters()).device

    def cutoff(self, xs, thresh=None):
        eps = torch.finfo(xs.dtype).eps
        
        if not thresh is None:
            if eps < thresh:
                eps = thresh

        xs = xs.clamp(min=eps)

        if torch.any(torch.isnan(xs)):
            xs[torch.isnan(xs)] = eps

        return xs

    def softmax(self, xs):
        #xs = SoftmaxTransform()(xs)
        xs = dist.Multinomial(total_count=1, logits=xs).mean
        return xs
    
    def sigmoid(self, xs):
        #sigm_enc = nn.Sigmoid()
        #xs = sigm_enc(xs)
        #xs = clamp_probs(xs)
        xs = dist.Bernoulli(logits=xs).mean
        return xs

    def softmax_logit(self, xs):
        eps = torch.finfo(xs.dtype).eps
        xs = self.softmax(xs)
        xs = torch.logit(xs, eps=eps)
        return xs

    def logit(self, xs):
        eps = torch.finfo(xs.dtype).eps
        xs = torch.logit(xs, eps=eps)
        return xs

    def dirimulti_param(self, xs):
        xs = self.dirimulti_mass * self.sigmoid(xs)
        return xs

    def multi_param(self, xs):
        xs = self.softmax(xs)
        return xs

    def model1(self, xs):
        pyro.module('sure', self)

        eps = torch.finfo(xs.dtype).eps
        batch_size = xs.size(0)
        self.options = dict(dtype=xs.dtype, device=xs.device)
        
        if self.loss_func=='negbinomial':
            total_count = pyro.param("inverse_dispersion", self.inverse_dispersion *
                                     xs.new_ones(self.input_size), constraint=constraints.positive)
            
        if self.use_zeroinflate:
            gate_logits = pyro.param("dropout_rate", xs.new_zeros(self.input_size))
            
        acs_scale = pyro.param("codebook_scale", xs.new_ones(self.latent_dim), constraint=constraints.positive)

        I = torch.eye(self.code_size)
        if self.latent_dist=='studentt':
            acs_dof,acs_loc = self.codebook(I)
        else:
            acs_loc = self.codebook(I)

        with pyro.plate('data'):
            prior = torch.zeros(batch_size, self.code_size, **self.options)
            ns = pyro.sample('n', dist.OneHotCategorical(logits=prior))

            zn_loc = torch.matmul(ns,acs_loc)
            #zn_scale = torch.matmul(ns,acs_scale)
            zn_scale = acs_scale

            if self.latent_dist == 'studentt':
                prior_dof = torch.matmul(ns,acs_dof)
                zns = pyro.sample('zn', dist.StudentT(df=prior_dof, loc=zn_loc, scale=zn_scale).to_event(1))
            elif self.latent_dist == 'laplacian':
                zns = pyro.sample('zn', dist.Laplace(zn_loc, zn_scale).to_event(1))
            elif self.latent_dist == 'cauchy':
                zns = pyro.sample('zn', dist.Cauchy(zn_loc, zn_scale).to_event(1))
            elif self.latent_dist == 'normal':
                zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))
            elif self.latent_dist == 'gumbel':
                zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1))

            zs = zns
            concentrate = self.decoder_concentrate(zs)
            rate = concentrate.exp()
            if self.loss_func != 'poisson':
                theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean

            if self.loss_func == 'negbinomial':
                if self.use_zeroinflate:
                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.NegativeBinomial(total_count=total_count, probs=theta),gate_logits=gate_logits).to_event(1), obs=xs)
                else:
                    pyro.sample('x', dist.NegativeBinomial(total_count=total_count, probs=theta).to_event(1), obs=xs)
            elif self.loss_func == 'poisson':
                if self.use_zeroinflate:
                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.Poisson(rate=rate),gate_logits=gate_logits).to_event(1), obs=xs.round())
                else:
                    pyro.sample('x', dist.Poisson(rate=rate).to_event(1), obs=xs.round())
            elif self.loss_func == 'multinomial':
                pyro.sample('x', dist.Multinomial(total_count=int(1e8), probs=theta), obs=xs)

    def guide1(self, xs):
        with pyro.plate('data'):
            zn_loc, zn_scale = self.encoder_zn(xs)
            zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))

            alpha = self.encoder_n(zns)
            ns = pyro.sample('n', dist.OneHotCategorical(logits=alpha))

    def model2(self, xs, us=None):
        pyro.module('sure', self)

        eps = torch.finfo(xs.dtype).eps
        batch_size = xs.size(0)
        self.options = dict(dtype=xs.dtype, device=xs.device)
        
        if self.loss_func=='negbinomial':
            total_count = pyro.param("inverse_dispersion", self.inverse_dispersion *
                                     xs.new_ones(self.input_size), constraint=constraints.positive)
            
        if self.use_zeroinflate:
            gate_logits = pyro.param("dropout_rate", xs.new_zeros(self.input_size))
            
        acs_scale = pyro.param("codebook_scale", xs.new_ones(self.latent_dim), constraint=constraints.positive)

        I = torch.eye(self.code_size)
        if self.latent_dist=='studentt':
            acs_dof,acs_loc = self.codebook(I)
        else:
            acs_loc = self.codebook(I)

        with pyro.plate('data'):
            prior = torch.zeros(batch_size, self.code_size, **self.options)
            ns = pyro.sample('n', dist.OneHotCategorical(logits=prior))

            zn_loc = torch.matmul(ns,acs_loc)
            #zn_scale = torch.matmul(ns,acs_scale)
            zn_scale = acs_scale

            if self.latent_dist == 'studentt':
                prior_dof = torch.matmul(ns,acs_dof)
                zns = pyro.sample('zn', dist.StudentT(df=prior_dof, loc=zn_loc, scale=zn_scale).to_event(1))
            elif self.latent_dist == 'laplacian':
                zns = pyro.sample('zn', dist.Laplace(zn_loc, zn_scale).to_event(1))
            elif self.latent_dist == 'cauchy':
                zns = pyro.sample('zn', dist.Cauchy(zn_loc, zn_scale).to_event(1))
            elif self.latent_dist == 'normal':
                zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))
            elif self.latent_dist == 'gumbel':
                zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1))

            if self.cell_factor_size>0:
                zus = self.cell_factor_effect(us)
                zs = zns + zus
            else:
                zs = zns

            concentrate = self.decoder_concentrate(zs)
            rate = concentrate.exp()
            if self.loss_func != 'poisson':
                theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean

            if self.loss_func == 'negbinomial':
                if self.use_zeroinflate:
                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.NegativeBinomial(total_count=total_count, probs=theta),gate_logits=gate_logits).to_event(1), obs=xs)
                else:
                    pyro.sample('x', dist.NegativeBinomial(total_count=total_count, probs=theta).to_event(1), obs=xs)
            elif self.loss_func == 'poisson':
                if self.use_zeroinflate:
                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.Poisson(rate=rate),gate_logits=gate_logits).to_event(1), obs=xs.round())
                else:
                    pyro.sample('x', dist.Poisson(rate=rate).to_event(1), obs=xs.round())
            elif self.loss_func == 'multinomial':
                pyro.sample('x', dist.Multinomial(total_count=int(1e8), probs=theta), obs=xs)

    def guide2(self, xs, us=None):
        with pyro.plate('data'):
            zn_loc, zn_scale = self.encoder_zn(xs)
            zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))

            alpha = self.encoder_n(zns)
            ns = pyro.sample('n', dist.OneHotCategorical(logits=alpha))

    def model3(self, xs, ys, embeds=None):
        pyro.module('sure', self)

        eps = torch.finfo(xs.dtype).eps
        batch_size = xs.size(0)
        self.options = dict(dtype=xs.dtype, device=xs.device)
        
        if self.loss_func=='negbinomial':
            total_count = pyro.param("inverse_dispersion", self.inverse_dispersion *
                                     xs.new_ones(self.input_size), constraint=constraints.positive)
            
        if self.use_zeroinflate:
            gate_logits = pyro.param("dropout_rate", xs.new_zeros(self.input_size))
            
        acs_scale = pyro.param("codebook_scale", xs.new_ones(self.latent_dim), constraint=constraints.positive)

        I = torch.eye(self.code_size)
        if self.latent_dist=='studentt':
            acs_dof,acs_loc = self.codebook(I)
        else:
            acs_loc = self.codebook(I)

        with pyro.plate('data'):
            #prior = torch.zeros(batch_size, self.code_size, **self.options)
            prior = self.encoder_n(xs)
            ns = pyro.sample('n', dist.OneHotCategorical(logits=prior), obs=ys)

            zn_loc = torch.matmul(ns,acs_loc)
            #prior_scale = torch.matmul(ns,acs_scale)
            zn_scale = acs_scale

            if self.latent_dist=='studentt':
                prior_dof = torch.matmul(ns,acs_dof)
                if embeds is None:
                    zns = pyro.sample('zn', dist.StudentT(df=prior_dof, loc=zn_loc, scale=zn_scale).to_event(1))
                else:
                    zns = pyro.sample('zn', dist.StudentT(df=prior_dof, loc=zn_loc, scale=zn_scale).to_event(1), obs=embeds)
            elif self.latent_dist=='laplacian':
                if embeds is None:
                    zns = pyro.sample('zn', dist.Laplace(zn_loc, zn_scale).to_event(1))
                else:
                    zns = pyro.sample('zn', dist.Laplace(zn_loc, zn_scale).to_event(1), obs=embeds)
            elif self.latent_dist=='cauchy':
                if embeds is None:
                    zns = pyro.sample('zn', dist.Cauchy(zn_loc, zn_scale).to_event(1))
                else:
                    zns = pyro.sample('zn', dist.Cauchy(zn_loc, zn_scale).to_event(1), obs=embeds)
            elif self.latent_dist=='normal':
                if embeds is None:
                    zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))
                else:
                    zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1), obs=embeds)
            elif self.z_dist == 'gumbel':
                if embeds is None:
                    zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1))
                else:
                    zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1), obs=embeds)

            zs = zns

            concentrate = self.decoder_concentrate(zs)
            rate = concentrate.exp()
            if self.loss_func != 'poisson':
                theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean

            if self.loss_func == 'negbinomial':
                if self.use_zeroinflate:
                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.NegativeBinomial(total_count=total_count, probs=theta),gate_logits=gate_logits).to_event(1), obs=xs)
                else:
                    pyro.sample('x', dist.NegativeBinomial(total_count=total_count, probs=theta).to_event(1), obs=xs)
            elif self.loss_func == 'poisson':
                if self.use_zeroinflate:
                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.Poisson(rate=rate),gate_logits=gate_logits).to_event(1), obs=xs.round())
                else:
                    pyro.sample('x', dist.Poisson(rate=rate).to_event(1), obs=xs.round())
            elif self.loss_func == 'multinomial':
                pyro.sample('x', dist.Multinomial(total_count=int(1e8), probs=theta), obs=xs)

    def guide3(self, xs, ys, embeds=None):
        with pyro.plate('data'):
            if embeds is None:
                zn_loc, zn_scale = self.encoder_zn(xs)
                zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))

    def model4(self, xs, us, ys, embeds=None):
        pyro.module('sure', self)

        eps = torch.finfo(xs.dtype).eps
        batch_size = xs.size(0)
        self.options = dict(dtype=xs.dtype, device=xs.device)
        
        if self.loss_func=='negbinomial':
            total_count = pyro.param("inverse_dispersion", self.inverse_dispersion *
                                     xs.new_ones(self.input_size), constraint=constraints.positive)
            
        if self.use_zeroinflate:
            gate_logits = pyro.param("dropout_rate", xs.new_zeros(self.input_size))
            
        acs_scale = pyro.param("codebook_scale", xs.new_ones(self.latent_dim), constraint=constraints.positive)

        I = torch.eye(self.code_size)
        if self.latent_dist=='studentt':
            acs_dof,acs_loc = self.codebook(I)
        else:
            acs_loc = self.codebook(I)

        with pyro.plate('data'):
            #prior = torch.zeros(batch_size, self.code_size, **self.options)
            prior = self.encoder_n(xs)
            ns = pyro.sample('n', dist.OneHotCategorical(logits=prior), obs=ys)

            zn_loc = torch.matmul(ns,acs_loc)
            #prior_scale = torch.matmul(ns,acs_scale)
            zn_scale = acs_scale

            if self.latent_dist=='studentt':
                prior_dof = torch.matmul(ns,acs_dof)
                if embeds is None:
                    zns = pyro.sample('zn', dist.StudentT(df=prior_dof, loc=zn_loc, scale=zn_scale).to_event(1))
                else:
                    zns = pyro.sample('zn', dist.StudentT(df=prior_dof, loc=zn_loc, scale=zn_scale).to_event(1), obs=embeds)
            elif self.latent_dist=='laplacian':
                if embeds is None:
                    zns = pyro.sample('zn', dist.Laplace(zn_loc, zn_scale).to_event(1))
                else:
                    zns = pyro.sample('zn', dist.Laplace(zn_loc, zn_scale).to_event(1), obs=embeds)
            elif self.latent_dist=='cauchy':
                if embeds is None:
                    zns = pyro.sample('zn', dist.Cauchy(zn_loc, zn_scale).to_event(1))
                else:
                    zns = pyro.sample('zn', dist.Cauchy(zn_loc, zn_scale).to_event(1), obs=embeds)
            elif self.latent_dist=='normal':
                if embeds is None:
                    zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))
                else:
                    zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1), obs=embeds)
            elif self.z_dist == 'gumbel':
                if embeds is None:
                    zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1))
                else:
                    zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1), obs=embeds)

            if self.cell_factor_size>0:
                zus = self.cell_factor_effect(us)
                zs = zus + zns
            else:
                zs = zns

            concentrate = self.decoder_concentrate(zs)
            rate = concentrate.exp()
            if self.loss_func != 'poisson':
                theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean

            if self.loss_func == 'negbinomial':
                if self.use_zeroinflate:
                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.NegativeBinomial(total_count=total_count, probs=theta),gate_logits=gate_logits).to_event(1), obs=xs)
                else:
                    pyro.sample('x', dist.NegativeBinomial(total_count=total_count, probs=theta).to_event(1), obs=xs)
            elif self.loss_func == 'poisson':
                if self.use_zeroinflate:
                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.Poisson(rate=rate),gate_logits=gate_logits).to_event(1), obs=xs.round())
                else:
                    pyro.sample('x', dist.Poisson(rate=rate).to_event(1), obs=xs.round())
            elif self.loss_func == 'multinomial':
                pyro.sample('x', dist.Multinomial(total_count=int(1e8), probs=theta), obs=xs)

    def guide4(self, xs, us, ys, embeds=None):
        with pyro.plate('data'):
            if embeds is None:
                zn_loc, zn_scale = self.encoder_zn(xs)
                zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))

    def _get_codebook(self):
        I = torch.eye(self.code_size, **self.options)
        if self.latent_dist=='studentt':
            _,cb = self.codebook(I)
        else:
            cb = self.codebook(I)
        return cb
    
    def get_codebook(self):
        """
        Return the mean part of metacell codebook
        """
        cb = self._get_metacell_coordinates()
        cb = tensor_to_numpy(cb)
        return cb

    def _get_cell_embedding(self, xs):           
        zns, _ = self.encoder_zn(xs)
        return zns 
    
    def get_cell_embedding(self, 
                             xs, 
                             batch_size: int = 1024):
        """
        Return cells' latent representations

        Parameters
        ----------
        xs
            Single-cell expression matrix. It should be a Numpy array or a Pytorch Tensor.
        batch_size
            Size of batch processing.
        use_decoder
            If toggled on, the latent representations will be reconstructed from the metacell codebook
        soft_assign
            If toggled on, the assignments of cells will use probabilistic values.
        """
        xs = self.preprocess(xs)
        xs = convert_to_tensor(xs, device=self.get_device())
        dataset = CustomDataset(xs)
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)

        Z = []
        with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
            for X_batch, _ in dataloader:
                zns = self._get_cell_embedding(X_batch)
                Z.append(tensor_to_numpy(zns))
                pbar.update(1)

        Z = np.concatenate(Z)
        return Z
    
    def _code(self, xs):
        if self.supervised_mode:
            alpha = self.encoder_n(xs)
        else:
            zns,_ = self.encoder_zn(xs)
            alpha = self.encoder_n(zns)
        return alpha
    
    def code(self, xs, batch_size=1024):
        xs = self.preprocess(xs)
        xs = convert_to_tensor(xs, device=self.get_device())
        dataset = CustomDataset(xs)
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)

        A = []
        with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
            for X_batch, _ in dataloader:
                a = self._code(X_batch)
                A.append(tensor_to_numpy(a))
                pbar.update(1)

        A = np.concatenate(A)
        return A
    
    def _soft_assignments(self, xs):
        alpha = self._code(xs)
        alpha = self.softmax(alpha)
        return alpha
    
    def soft_assignments(self, xs, batch_size=1024):
        """
        Map cells to metacells and return the probabilistic values of metacell assignments
        """
        xs = self.preprocess(xs)
        xs = convert_to_tensor(xs, device=self.get_device())
        dataset = CustomDataset(xs)
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)

        A = []
        with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
            for X_batch, _ in dataloader:
                a = self._soft_assignments(X_batch)
                A.append(tensor_to_numpy(a))
                pbar.update(1)

        A = np.concatenate(A)
        return A
    
    def _hard_assignments(self, xs):
        alpha = self._code(xs)
        res, ind = torch.topk(alpha, 1)
        ns = torch.zeros_like(alpha).scatter_(1, ind, 1.0)
        return ns
    
    def hard_assignments(self, xs, batch_size=1024):
        """
        Map cells to metacells and return the assigned metacell identities.
        """
        xs = self.preprocess(xs)
        xs = convert_to_tensor(xs, device=self.get_device())
        dataset = CustomDataset(xs)
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)

        A = []
        with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
            for X_batch, _ in dataloader:
                a = self._hard_assignments(X_batch)
                A.append(tensor_to_numpy(a))
                pbar.update(1)

        A = np.concatenate(A)
        return A
    
    def preprocess(self, xs):
        if sparse.issparse(xs):
            xs = xs.toarray()
        return xs 
    
    def fit(self, xs, 
            us = None, 
            ys = None,
            zs = None,
            num_epochs: int = 200, 
            learning_rate: float = 0.0001, 
            batch_size: int = 256, 
            algo: Literal['adam','rmsprop','adamw'] = 'adam', 
            beta_1: float = 0.9, 
            weight_decay: float = 0.005, 
            decay_rate: float = 0.9,
            config_enum: str = 'parallel',
            use_jax: bool = False):
        """
        Train the SURE model.

        Parameters
        ----------
        xs
            Single-cell experssion matrix. It should be a Numpy array or a Pytorch Tensor. Rows are cells and columns are features.
        us
            Undesired factor matrix. It should be a Numpy array or a Pytorch Tensor. Rows are cells and columns are undesired factors.
        ys
            Desired factor matrix. It should be a Numpy array or a Pytorch Tensor. Rows are cells and columns are desired factors.
        num_epochs
            Number of training epochs.
        learning_rate
            Parameter for training.
        batch_size
            Size of batch processing.
        algo
            Optimization algorithm.
        beta_1
            Parameter for optimization.
        weight_decay
            Parameter for optimization.
        decay_rate 
            Parameter for optimization.
        use_jax
            If toggled on, Jax will be used for speeding up. CAUTION: This will raise errors because of unknown reasons when it is called in
            the Python script or Jupyter notebook. It is OK if it is used when runing SURE in the shell command.
        """
        xs = self.preprocess(xs)
        xs = convert_to_tensor(xs, dtype=self.dtype, device=self.get_device())
        if us is not None:
            us = convert_to_tensor(us, dtype=self.dtype, device=self.get_device())
        if ys is not None:
            ys = convert_to_tensor(ys, dtype=self.dtype, device=self.get_device())
        if zs is not None:
            zs = convert_to_tensor(zs, dtype=self.dtype, device=self.get_device())

        dataset = CustomDataset4(xs, us, ys, zs)
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True)

        # setup the optimizer
        optim_params = {'lr': learning_rate, 'betas': (beta_1, 0.999), 'weight_decay': weight_decay}

        if algo.lower()=='rmsprop':
            optimizer = torch.optim.RMSprop
        elif algo.lower()=='adam':
            optimizer = torch.optim.Adam
        elif algo.lower() == 'adamw':
            optimizer = torch.optim.AdamW
        else:
            raise ValueError("An optimization algorithm must be specified.")
        scheduler = ExponentialLR({'optimizer': optimizer, 'optim_args': optim_params, 'gamma': decay_rate})

        pyro.clear_param_store()

        # set up the loss(es) for inference, wrapping the guide in config_enumerate builds the loss as a sum
        # by enumerating each class label form the sampled discrete categorical distribution in the model
        Elbo = JitTraceEnum_ELBO if use_jax else TraceEnum_ELBO
        elbo = Elbo(max_plate_nesting=1, strict_enumeration_warning=False)
        if us is None:
            if ys is None:
                guide = config_enumerate(self.guide1, config_enum, expand=True)
                loss_basic = SVI(self.model1, guide, scheduler, loss=elbo)
            else:
                guide = config_enumerate(self.guide3, config_enum, expand=True)
                loss_basic = SVI(self.model3, guide, scheduler, loss=elbo)
        else:
            if ys is None:
                guide = config_enumerate(self.guide2, config_enum, expand=True)
                loss_basic = SVI(self.model2, guide, scheduler, loss=elbo)
            else:
                guide = config_enumerate(self.guide4, config_enum, expand=True)
                loss_basic = SVI(self.model4, guide, scheduler, loss=elbo)

        # build a list of all losses considered
        losses = [loss_basic]
        num_losses = len(losses)

        with tqdm(total=num_epochs, desc='Training', unit='epoch') as pbar:
            for epoch in range(num_epochs):
                epoch_losses = [0.0] * num_losses
                for batch_x, batch_u, batch_y, batch_z, _ in dataloader:
                    if us is None:
                        batch_u = None
                    if ys is None:
                        batch_y = None
                    if zs is None:
                        batch_z = None

                    for loss_id in range(num_losses):
                        if batch_u is None:
                            if batch_y is None:
                                new_loss = losses[loss_id].step(batch_x)
                            else:
                                new_loss = losses[loss_id].step(batch_x, batch_y, batch_z)
                        else:
                            if batch_y is None:
                                new_loss = losses[loss_id].step(batch_x, batch_u)
                            else:
                                new_loss = losses[loss_id].step(batch_x, batch_u, batch_y, batch_z)
                        epoch_losses[loss_id] += new_loss

                avg_epoch_losses_ = map(lambda v: v / len(dataloader), epoch_losses)
                avg_epoch_losses = map(lambda v: "{:.4f}".format(v), avg_epoch_losses_)

                # store the loss
                str_loss = " ".join(map(str, avg_epoch_losses))

                # Update progress bar
                pbar.set_postfix({'loss': str_loss})
                pbar.update(1)
        
        assigns = self.soft_assignments(xs)
        assigns = convert_to_tensor(assigns, dtype=self.dtype, device=self.get_device())
        self.codebook_weights = torch.sum(assigns, dim=0)
        self.codebook_weights = self.codebook_weights / torch.sum(self.codebook_weights)
        #self.inverse_dispersion = pyro.param('inverse_dispersion').item()
        #self.dof = pyro.param('dof').item()

    @classmethod
    def save_model(cls, model, file_path, compression=False):
        """Save the model to the specified file path."""
        file_path = os.path.abspath(file_path)

        model.eval()
        if compression:
            with gzip.open(file_path, 'wb') as pickle_file:
                pickle.dump(model, pickle_file)
        else:
            with open(file_path, 'wb') as pickle_file:
                pickle.dump(model, pickle_file)

        print(f'Model saved to {file_path}')

    @classmethod
    def load_model(cls, file_path):
        """Load the model from the specified file path and return an instance."""
        print(f'Model loaded from {file_path}')

        file_path = os.path.abspath(file_path)
        if file_path.endswith('gz'):
            with gzip.open(file_path, 'rb') as pickle_file:
                model = pickle.load(pickle_file)
        else:
            with open(file_path, 'rb') as pickle_file:
                model = pickle.load(pickle_file)
        
        return model

        
EXAMPLE_RUN = (
    "example run: SURE --help"
)

def parse_args():
    parser = argparse.ArgumentParser(
        description="SURE\n{}".format(EXAMPLE_RUN))

    parser.add_argument(
        "--cuda", action="store_true", help="use GPU(s) to speed up training"
    )
    parser.add_argument(
        "--jit", action="store_true", help="use PyTorch jit to speed up training"
    )
    parser.add_argument(
        "-n", "--num-epochs", default=200, type=int, help="number of epochs to run"
    )
    parser.add_argument(
        "-enum",
        "--enum-discrete",
        default="parallel",
        help="parallel, sequential or none. uses parallel enumeration by default",
    )
    parser.add_argument(
        "-data",
        "--data-file",
        default=None,
        type=str,
        help="the data file",
    )
    parser.add_argument(
        "-undesired",
        "--undesired-factor-file",
        default=None,
        type=str,
        help="the file for the record of undesired factors",
    )
    parser.add_argument(
        "-delta",
        "--delta",
        default=0.0,
        type=float,
        help="penalty weight for zero inflation loss",
    )
    parser.add_argument(
        "-64",
        "--float64",
        action="store_true",
        help="use double float precision",
    )
    parser.add_argument(
        "--z-dist",
        default='normal',
        type=str,
        choices=['normal','laplacian','studentt','cauchy'],
        help="distribution model for latent representation",
    )
    parser.add_argument(
        "-cs",
        "--codebook-size",
        default=100,
        type=int,
        help="size of vector quantization codebook",
    )
    parser.add_argument(
        "-zd",
        "--z-dim",
        default=10,
        type=int,
        help="size of the tensor representing the latent variable z variable",
    )
    parser.add_argument(
        "-hl",
        "--hidden-layers",
        nargs="+",
        default=[500],
        type=int,
        help="a tuple (or list) of MLP layers to be used in the neural networks "
        "representing the parameters of the distributions in our model",
    )
    parser.add_argument(
        "-hla",
        "--hidden-layer-activation",
        default='relu',
        type=str,
        choices=['relu','softplus','leakyrelu','linear'],
        help="activation function for hidden layers",
    )
    parser.add_argument(
        "-plf",
        "--post-layer-function",
        nargs="+",
        default=['layernorm'],
        type=str,
        help="post functions for hidden layers, could be none, dropout, layernorm, batchnorm, or combination, default is 'dropout layernorm'",
    )
    parser.add_argument(
        "-paf",
        "--post-activation-function",
        nargs="+",
        default=['none'],
        type=str,
        help="post functions for activation layers, could be none or dropout, default is 'none'",
    )
    parser.add_argument(
        "-id",
        "--inverse-dispersion",
        default=10.0,
        type=float,
        help="inverse dispersion prior for negative binomial",
    )
    parser.add_argument(
        "-lr",
        "--learning-rate",
        default=0.0001,
        type=float,
        help="learning rate for Adam optimizer",
    )
    parser.add_argument(
        "-dr",
        "--decay-rate",
        default=0.9,
        type=float,
        help="decay rate for Adam optimizer",
    )
    parser.add_argument(
        "--layer-dropout-rate",
        default=0.1,
        type=float,
        help="droput rate for neural networks",
    )
    parser.add_argument(
        "-b1",
        "--beta-1",
        default=0.95,
        type=float,
        help="beta-1 parameter for Adam optimizer",
    )
    parser.add_argument(
        "-bs",
        "--batch-size",
        default=1000,
        type=int,
        help="number of cells to be considered in a batch",
    )
    parser.add_argument(
        "-gp",
        "--gate-prior",
        default=0.6,
        type=float,
        help="gate prior for zero-inflated model",
    )
    parser.add_argument(
        "-likeli",
        "--likelihood",
        default='negbinomial',
        type=str,
        choices=['negbinomial', 'multinomial', 'poisson', 'gaussian','lognormal'],
        help="specify the distribution likelihood function",
    )
    parser.add_argument(
        "-dirichlet",
        "--use-dirichlet",
        action="store_true",
        help="use Dirichlet distribution over gene frequency",
    )
    parser.add_argument(
        "-mass",
        "--dirichlet-mass",
        default=1,
        type=float,
        help="mass param for dirichlet model",
    )
    parser.add_argument(
        "-zi",
        "--zero-inflation",
        default='exact',
        type=str,
        choices=['none','exact','inexact'],
        help="use zero-inflated estimation",
    )
    parser.add_argument(
        "--seed",
        default=None,
        type=int,
        help="seed for controlling randomness in this example",
    )
    parser.add_argument(
        "--save-model",
        default=None,
        type=str,
        help="path to save model for prediction",
    )
    args = parser.parse_args()
    return args

def main():
    args = parse_args()
    assert (
        (args.data_file is not None) and (
            os.path.exists(args.data_file))
    ), "data file must be provided"

    if args.seed is not None:
        set_random_seed(args.seed)

    if args.float64:
        dtype = torch.float64
        torch.set_default_dtype(torch.float64)
    else:
        dtype = torch.float32
        torch.set_default_dtype(torch.float32)

    xs = dt.fread(file=args.data_file, header=True).to_numpy()
    us = None 
    if args.undesired_factor_file is not None:
        us = dt.fread(file=args.undesired_factor_file, header=True).to_numpy()

    input_size = xs.shape[1]
    undesired_size = 0 if us is None else us.shape[1]

    latent_dist = args.z_dist

    ###########################################
    sure = SURE(
        input_size=input_size,
        undesired_size=undesired_size,
        inverse_dispersion=args.inverse_dispersion,
        latent_dim=args.latent_dim,
        hidden_layers=args.hidden_layers,
        hidden_layer_activation=args.hidden_layer_activation,
        use_cuda=args.cuda,
        config_enum=args.enum_discrete,
        use_dirichlet=args.use_dirichlet,
        zero_inflation=args.zero_inflation,
        gate_prior=args.gate_prior,
        delta=args.delta,
        loss_func=args.likelihood,
        dirichlet_mass=args.dirichlet_mass,
        nn_dropout=args.layer_dropout_rate,
        post_layer_fct=args.post_layer_function,
        post_act_fct=args.post_activation_function,
        codebook_size=args.codebook_size,
        latent_dist = latent_dist,
        dtype=dtype,
    )

    sure.fit(xs, us=us, 
             num_epochs=args.num_epochs,
             learning_rate=args.learning_rate,
             batch_size=args.batch_size,
             beta_1=args.beta_1,
             decay_rate=args.decay_rate,
             use_jax=args.jit,
             config_enum=args.enum_discrete,
             )

    if args.save_model is not None:
        if args.save_model.endswith('gz'):
            SURE.save_model(sure, args.save_model, compression=True)
        else:
            SURE.save_model(sure, args.save_model)
    


if __name__ == "__main__":

    main()


import os 
import tempfile
import subprocess 
import scanpy as sc
import numpy as np
import scipy as sp
from scipy import sparse
from scipy.stats import gaussian_kde

if sp.__version__ < '1.14.0':
    from scipy.integrate import cumtrapz
else:
    from scipy.integrate import cumulative_trapezoid as cumtrapz

import pandas as pd
from functools import reduce

import datatable as dt
from tqdm import tqdm
import umap 
import faiss 
from sklearn.neighbors import NearestNeighbors
#from cuml.linear_model import LogisticRegression
from sklearn.preprocessing import LabelEncoder,LabelBinarizer

import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import pyro.distributions as dist

from ..SURE import SURE
from .assembly import assembly,get_data,get_subdata,batch_encoding,get_uns
from ..codebook import codebook_summarize_,codebook_generate,codebook_sketch
from ..utils import convert_to_tensor, tensor_to_numpy
from ..utils import CustomDataset
from ..utils import pretty_print, Colors
from ..utils import PriorityQueue

import networkx as nx
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.colors import ListedColormap

import dill as pickle
import gzip 
from packaging.version import Version
torch_version = torch.__version__

from typing import Literal

import warnings
warnings.filterwarnings("ignore")

class SingleOmicsAtlas(nn.Module):
    """
    Compressed Cell Atlas

    Parameters
    ----------
    atlas_name
        Name of the built atlas.
    hvgs
        Highly variable genes.
    eps
        Low bound.
    """
    def __init__(self, 
                 atlas_name: str = 'Atlas', 
                 hvgs: list = None, 
                 eps: float = 1e-12):
        super().__init__()
        self.atlas_name = atlas_name
        self.model = None 
        self.sure_models_list = None 
        self.hvgs = hvgs 
        self.adata = None
        self.sample_adata = None
        self.layer = None
        self.n_sure_models = None
        self.umap_metric='euclidean'
        self.umap = None
        self.adj = None
        self.subatlas_list = None 
        self.n_subatlas = 0
        self.pheno_keys = None
        self.nearest_neighbor_engine = None
        self.knn_k = 5
        self.network = None
        self.network_pos = None
        self.sample_network = None 
        self.sample_network_pos = None
        self.eps=eps

    def fit(self, adata_list_, 
            batch_key: str = None, 
            pheno_keys: list = None, 
            preprocessing: bool = True, 
            hvgs: list = None,
            n_top_genes: int = 5000, 
            hvg_method: Literal['seurat','seurat_v3','cell_ranger'] ='seurat', 
            layer: str = 'counts', 
            cuda_id: int = 0, 
            use_jax: bool = False,
            codebook_size: int = 800, 
            codebook_size_per_adata: int = 500, 
            min_gamma=20,
            min_cell=1000,
            learning_rate: float = 0.0001,
            batch_size: int = 200, 
            batch_size_per_adata: int = 100, 
            n_epochs: int = 200, 
            latent_dist: Literal['normal','laplacian','studentt','cauchy'] = 'studentt',
            use_dirichlet: bool = False, 
            use_dirichlet_per_adata: bool = False, 
            zero_inflation: bool = False, 
            zero_inflation_per_adata: bool = False,
            likelihood: Literal['negbinomial','poisson','multinomial','gaussian'] = 'multinomial', 
            likelihood_per_adata: Literal['negbinomial','poisson','multinomial','gaussian'] = 'multinomial', 
            #n_samples_per_adata: int = 10000, 
            #total_samples: int = 500000, 
            #summarize: bool = True,
            sketch_func: Literal['mean','sum','simulate','sample','bootstrap_sample'] = 'mean',
            n_bootstrap_neighbors: int = 8,
            #sketching: bool = True, 
            #even_sketch: bool = True,
            #bootstrap_sketch: bool = False, 
            #n_sketch_neighbors: int = 10, 
            n_workers: int = 1,
            mute: bool = True,
            edge_thresh: float = 0.001,
            metric: Literal['euclidean','correlation','cosine'] ='euclidean',
            knn_k: int = 5):
        """
        Fit the input list of AnnData datasets.

        Parameters
        ----------
        adata_list
            A list of AnnData datasets.
        batch_key
            Undesired factor. 
        pheno_keys
            A list of phenotype factors, of which the information should be retained in the built atlas.
        preprocessing
            If toggled on, the input datasets will go through the standard Scanpy proprocessing steps including normalization and log1p transformation.
        hvgs
            If a list of highly variable genes is given, the subsequent steps will rely on these genes.
        n_top_genes
            Parameter for Scanpy's highly_variable_genes
        hvg_method
            Parameter for Scanpy's highly_variable_genes
        layer
            Data used for building the atlas.
        cuda_id
            Cuda device.
        use_jax
            If toggled on, Jax will be used for speeding.
        codebook_size
            Size of metacells in the built atlas.
        codebook_size_per_adata
            Size of metacells for each adata.
        learning_rate
            Parameter for optimization.
        batch_size
            Parameter for building the atlas.
        batch_size_per_adata
            Parameter for calling metacells within each adata.
        n_epochs
            Number of epochs.
        latent_dist
            Distribution for latent representations.
        use_dirichlet
            Use Dirichlet model for building the atlas.
        use_dirichlet_per_adata
            Use Dirichlet model for calling metacells within each adata.
        zero_inflation
            Use zero-inflated model for building the atlas.
        zero_inflation_per_adata
            Use zero-inflated model for calling metacells within each adata.
        likelihood
            Data generation model for building the atlas.
        likelihood_per_adata
            Data generation model for calling metacells within each adata.
        n_samples_per_adata
            Number of samples drawn from each adata for building the atlas.
        total_samples
            Total number of samples for building the atlas.
        sketching
            If toggled on, sketched cells will be used for building the atlas.
        bootstrap_sketch
            If toggled on, bootstraped sketching will be used instead of simple sketching.
        n_sketch_neighbors
            Parameter for bootstraped sketching.
        edge_thresh
            Parameter for building network.
        metric
            Parameter for UMAP.
        knn_k
            Parameter for K-nearest-neighbor machine.
        """
        
        print(Colors.YELLOW + 'Create A Distribution-Preserved Single-Cell Omics Atlas' + Colors.RESET)
        adata_list = [ad for ad in adata_list_ if ad.shape[0]>min_cell]
        n_adatas = len(adata_list)
        self.layer = layer
        self.n_sure_models = n_adatas
        self.umap_metric = metric
        self.pheno_keys = pheno_keys
        #zero_inflation = True if sketching else zero_inflation

        # assembly
        print(f'{n_adatas} adata datasets are given')
        self.model,self.submodels,self.hvgs = assembly(adata_list, batch_key, 
                 preprocessing, hvgs, n_top_genes, hvg_method, layer, cuda_id, use_jax,
                 codebook_size, codebook_size_per_adata, min_gamma, learning_rate,
                 batch_size, batch_size_per_adata, n_epochs, latent_dist,
                 use_dirichlet, use_dirichlet_per_adata,
                 zero_inflation, zero_inflation_per_adata,
                 likelihood, likelihood_per_adata,
                 #n_samples_per_adata, total_samples, summarize,
                 sketch_func,n_bootstrap_neighbors,
                 #sketching, even_sketch, bootstrap_sketch, n_sketch_neighbors, 
                 n_workers, mute)
        
        # summarize expression
        X,W,adj = None,None,None
        with tqdm(total=n_adatas, desc=f'Summarize data in {layer}', unit='adata') as pbar:
            for i in np.arange(n_adatas):
                #print(f'Adata {i+1} / {n_adatas}: Summarize data in {layer}')
                adata_i = adata_list[i][:,self.hvgs].copy()
                adata_i_ = adata_list[i].copy()

                xs_i = get_data(adata_i, layer).values
                xs_i_ = get_data(adata_i_, layer).values
                ws_i_sup = self.model.soft_assignments(xs_i)
                xs_i_sup = codebook_summarize_(ws_i_sup, xs_i_)

                if X is None:
                    X = xs_i_sup
                    W = np.sum(ws_i_sup.T, axis=1, keepdims=True)

                    a = convert_to_tensor(ws_i_sup)
                    a_t = a.T / torch.sum(a.T, dim=1, keepdim=True)
                    adj = torch.matmul(a_t, a)
                else:
                    X += xs_i_sup
                    W += np.sum(ws_i_sup.T, axis=1, keepdims=True)

                    a = convert_to_tensor(ws_i_sup)
                    a_t = a.T / torch.sum(a.T, dim=1, keepdim=True)
                    adj += torch.matmul(a_t, a)
                    
                pbar.update(1)
        X = X / W 
        self.adata = sc.AnnData(X)
        self.adata.obs_names = [f'MC{x}' for x in self.adata.obs_names]
        self.adata.var_names = adata_i_.var_names

        adj = tensor_to_numpy(adj) / self.n_sure_models
        self.adj = (adj + adj.T) / 2
        n_nodes = adj.shape[0]
        self.adj[np.arange(n_nodes), np.arange(n_nodes)] = 0

        # summarize phenotypes
        if pheno_keys is not None:
            self._summarize_phenotypes_from_adatas(adata_list, pheno_keys)

        # COMMENT OUT 2025.7.4
        # compute visualization position for the atlas
        #print('Compute the reference position of the atlas')
        #n_samples = np.max([n_samples_per_adata * self.n_sure_models, 50000])
        #n_samples = np.min([n_samples, total_samples])
        #self.instantiation(n_samples)
        #
        # create nearest neighbor indexing
        #self.build_nearest_neighbor_engine(knn_k)
        #self.knn_k = knn_k
        #
        #self.build_network(edge_thresh=edge_thresh)
        # END OF COMMENT OUT 2025.7.4
        
        # the distribution of cell-to-metacell distances
        metacells = self.model.get_metacell_coordinates()
        self.nearest_metacell_engine = NearestNeighbors(n_neighbors=knn_k,n_jobs=-1)
        self.nearest_metacell_engine.fit(metacells)
        
        cell2metacell_distances = []
        with tqdm(total=n_adatas, desc='Build cell-to-metacell distance distribution', unit='adata') as pbar:
            for i in np.arange(n_adatas):
                #print(f'Adata {i+1} / {n_adatas}: Build cell-to-metacell distance distribution')
                adata_i = adata_list[i][:,self.hvgs].copy()

                xs_i = get_data(adata_i, layer).values
                zs_i = self.model.get_cell_coordinates(xs_i)

                dd_i,_ = self.nearest_metacell_engine.kneighbors(zs_i, n_neighbors=1)
                cell2metacell_distances.extend(dd_i.flatten())
                
                pbar.update(1)
            
        self.cell2metacell_dist = gaussian_kde(cell2metacell_distances)
        
        print(Colors.YELLOW + f'A distribution-preserved atlas has been built from {n_adatas} adata datasets.' + Colors.RESET)
        
    def detect_outlier(self, adata_query, thresh:float = 1e-2, batch_size:int = 1024):
        
        def batch_p_values_greater(kde, x_vector):
            """
            批量计算向量x中每个元素的P(X > x)
            
            参数:
                kde: 已拟合的gaussian_kde对象
                x_vector: 包含多个x值的数组
                
            返回:
                与x_vector形状相同的P(X > x)数组
            """
            # 创建高分辨率的CDF查找表
            x_min = min(kde.dataset.min(), x_vector.min()) - 3 * np.std(kde.dataset)
            x_max = max(kde.dataset.max(), x_vector.max()) + 3 * np.std(kde.dataset)
            grid_points = max(20000, len(kde.dataset))  # 确保足够密集
            x_grid = np.linspace(x_min, x_max, grid_points)
            
            # 计算PDF和CDF
            pdf = kde(x_grid)
            cdf = np.cumsum(pdf)
            cdf /= cdf[-1]  # 归一化
            
            # 使用插值查找每个x对应的P(X > x) = 1 - CDF(x)
            p_values = 1 - np.interp(x_vector, x_grid, cdf)
            
            # 处理边界外的值
            p_values[x_vector < x_min] = 1.0
            p_values[x_vector > x_max] = 0.0
            
            return p_values
        
        adata_query = adata_query.copy()
        X_query = get_subdata(adata_query, self.hvgs, self.layer).values
        Z_map = self.model.get_cell_coordinates(X_query, batch_size=batch_size)
        
        dd,_ = self.nearest_metacell_engine.kneighbors(Z_map, n_neighbors=1)
        #pp = self.cell2metacell_dist(dd.flatten())
        pp = batch_p_values_greater(self.cell2metacell_dist,dd)
        outliers = pp < thresh
        
        print(f'{np.sum(outliers)} outliers found.')
        
        return outliers
        
    def map(self, adata_query, 
            batch_size: int = 1024):
        """
        Map query data to the atlas.

        Parameters
        ----------
        adata_query
            Query data. It should be an AnnData object.
        batch_size
            Size of batch processing.
        """
        adata_query = adata_query.copy()
        X_query = get_subdata(adata_query, self.hvgs, self.layer).values

        Z_map = self.model.get_cell_coordinates(X_query, batch_size=batch_size)
        A_map = self.model.soft_assignments(X_query)

        return Z_map, A_map
    
    def phenotype_density_estimation(self, adata_list, pheno_key):
        n_adatas = len(adata_list)
        PH_MC_list = []
        with tqdm(total=n_adatas, desc=f'Estimate {pheno_key} density', unit='adata') as pbar:
            for i in np.arange(n_adatas):
                ad = adata_list[i][:,self.hvgs].copy()
                xs = get_data(ad, layer=self.layer).values
                X_MC_mat = self.model.soft_assignments(xs)
                lb = LabelBinarizer().fit(ad.obs[pheno_key].astype(str))
                X_PH_mat = lb.transform(ad.obs[pheno_key].astype(str))
                PH_MC_mat = np.matmul(X_PH_mat.T, X_MC_mat)
                
                p_PH = np.sum(PH_MC_mat, axis=1, keepdims=True)
                p_PH[p_PH==0] = 1
                p_MC_PH = PH_MC_mat / p_PH
                PH_MC_mat = p_MC_PH * p_PH
                p_MC = np.sum(PH_MC_mat, axis=0, keepdims=True)
                p_MC[p_MC==0] = 1
                p_PH_MC = PH_MC_mat / p_MC
                
                PH_MC_df = pd.DataFrame(p_PH_MC, 
                                        columns = [f'MC{x}' for x in np.arange(X_MC_mat.shape[1])],
                                        index = lb.classes_)
                PH_MC_list.append(PH_MC_df)
                pbar.update(1)
        
        if n_adatas>1:
            PH_MC_DF = aggregate_dataframes(PH_MC_list)
            PH_MC_DF = PH_MC_DF.div(n_adatas)
        else:
            PH_MC_DF = PH_MC_list[0]
            
        self.adata.uns[f'{pheno_key}_density'] = PH_MC_DF
    
    def summarize_phenotypes(self, adata_list=None, pheno_keys=None):
        self._summarize_phenotypes_from_adatas(adata_list, pheno_keys)

    def _summarize_phenotypes_from_adatas(self, adata_list, pheno_keys):
        n_adatas = len(adata_list)
        for pheno in pheno_keys:
            Y = list()
            with tqdm(total=n_adatas, desc=f'Summarize data in {pheno}', unit='adata') as pbar:
                for i in np.arange(n_adatas):
                    if pheno in adata_list[i].obs.columns:
                        #print(f'Adata {i+1} / {n_adatas}: Summarize data in {pheno}')
                        adata_i = adata_list[i][:,self.hvgs].copy()

                        xs_i = get_data(adata_i, self.layer).values
                        ws_i_sup = self.model.soft_assignments(xs_i)
                        #ys_i = batch_encoding(adata_i, pheno)
                        #columns_i = ys_i.columns.tolist()
                        #ys_i = codebook_summarize_(ws_i_sup, ys_i.values)
                    
                        ws_i = np.argmax(ws_i_sup, axis=1)
                        adata_i.obs['metacell'] = [f'MC{x}' for x in ws_i]
                        df = adata_i.obs[['metacell',pheno]].value_counts().unstack(fill_value=0)

                        Y.append(df)
                        pbar.update(1)

            Y_df_ = aggregate_dataframes(Y)
            Y_df = pd.DataFrame(0, index=[f'MC{x}' for x in np.arange(self.adata.shape[0])], columns=Y_df_.columns)
            Y_df = Y_df.add(Y_df_, fill_value=0)
            
            #Y = Y_df.values
            #Y[Y<self.eps] = 0
            #Y = Y / Y.sum(axis=1, keepdims=True)

            self.adata.uns[pheno] = Y_df
            #self.adata.uns[f'{pheno}_columns'] = Y_df.columns
            Y_hat_ = Y_df.idxmax(axis=1)
            Y_hat = pd.DataFrame(pd.NA, index=[f'MC{x}' for x in np.arange(self.adata.shape[0])], columns=['id'])
            Y_hat.loc[Y_hat_.index.tolist(),'id'] = Y_hat_.tolist()
            self.adata.obs[pheno] = Y_hat['id'].tolist()

    def phenotype_predict(self, adata_query, pheno_key, batch_size=1024):
        _,ws = self.map(adata_query, batch_size)
        A = matrix_dotprod(ws, self.adata.uns[f'{pheno_key}_density'].values)
        A = pd.DataFrame(A, columns=self.adata.uns[f'{pheno_key}_density'].columns)
        return A.idxmax(axis=1).tolist()
    
    @classmethod
    def save_model(cls, atlas, file_path, compression=False):
        """Save the model to the specified file path."""
        file_path = os.path.abspath(file_path)

        atlas.sample_adata = None
        atlas.eval()

        if compression:
            with gzip.open(file_path, 'wb') as pickle_file:
                pickle.dump(atlas, pickle_file)
        else:
            with open(file_path, 'wb') as pickle_file:
                pickle.dump(atlas, pickle_file)

        print(f'Model saved to {file_path}')

    @classmethod
    def load_model(cls, file_path, n_samples=10000):
        """Load the model from the specified file path and return an instance."""
        print(f'Model loaded from {file_path}')

        file_path = os.path.abspath(file_path)
        if file_path.endswith('gz'):
            with gzip.open(file_path, 'rb') as pickle_file:
                atlas = pickle.load(pickle_file)
        else:
            with open(file_path, 'rb') as pickle_file:
                atlas = pickle.load(pickle_file)
        
        #xs = atlas.sample(n_samples)
        #atlas.sample_adata = sc.AnnData(xs)
        #atlas.sample_adata.var_names = atlas.hvgs
        #
        #zs = atlas.model.get_cell_coordinates(xs)
        #ws = atlas.model.soft_assignments(xs)
        #atlas.sample_adata.obsm['X_umap'] = atlas.umap.transform(zs)
        #atlas.sample_adata.obsm['X_sure'] = zs 
        #atlas.sample_adata.obsm['weight'] = ws
        #
        return atlas




#def aggregate_dataframes(df_list):
#    n_dfs = len(df_list)
#    all_columns = set(df_list[0].columns)
#    for i in np.arange(n_dfs-1):
#        all_columns = all_columns.union(set(df_list[i+1].columns))
#        
#    all_indexs = set(df_list[0].index.tolist())
#    for i in np.arange(n_dfs-1):
#        all_indexs = all_indexs.union(set(df_list[i+1].index.tolist()))
#        
#    for col in all_columns:
#        for i in np.arange(n_dfs):
#            if col not in df_list[i]:
#                df_list[i][col] = 0  
#
#    df = pd.DataFrame(0, index=all_indexs, columns=all_columns)
#    df = df_list[0]
#    for i in np.arange(n_dfs-1):
#        df += df_list[i+1]
#
#    #df /= n_dfs
#    return df

def aggregate_dataframes(df_list):
    all_index = reduce(lambda x, y: x.union(y), [df.index for df in df_list], pd.Index([]))
    all_index_sorted = all_index.sort_values()
    all_columns = reduce(lambda x, y: x.union(y), [df.columns for df in df_list], pd.Index([]))
    result = pd.DataFrame(0, index=all_index_sorted, columns=all_columns)
    
    for df in df_list:
        result = result.add(df, fill_value=0)
        
    return result

def smooth_y_over_x(xs, ys, knn_k):
    n = xs.shape[0]
    nbrs = NearestNeighbors(n_neighbors=knn_k, n_jobs=-1)
    nbrs.fit(xs)
    ids = nbrs.kneighbors(xs, return_distance=False)
    ys_smooth = np.zeros_like(ys)
    for i in np.arange(knn_k):
        ys_smooth += ys[ids[:,i]]
    ys_smooth -= ys
    ys_smooth /= knn_k-1
    return ys_smooth

def matrix_dotprod(A, B, dtype=torch.float32):
    A = convert_to_tensor(A, dtype=dtype)
    B = convert_to_tensor(B, dtype=dtype)
    AB = torch.matmul(A, B)
    return tensor_to_numpy(AB)

def matrix_elemprod(A, B):
    A = convert_to_tensor(A)
    B = convert_to_tensor(B)
    AB = A * B
    return tensor_to_numpy(AB)

def cdf(density, xs, initial=0):
    CDF = cumtrapz(density, xs, initial=initial)
    CDF /= CDF[-1]
    return CDF



class FaissKNeighbors:
    def __init__(self, n_neighbors=5):
        self.index = None
        self.k = n_neighbors

    def fit(self, X):
        self.index = faiss.IndexFlatL2(X.shape[1])
        self.index.add(X.astype(np.float32))

    def kneighbors(self, X):
        distances, indices = self.index.search(X.astype(np.float32), k=self.k)
        return distances, indices
    




import torch
from torch.utils.data import DataLoader

import pyro
import pyro.distributions as dist

import numpy as np
import scipy as sp
import pandas as pd
from scipy.stats import gaussian_kde
from sklearn.preprocessing import LabelBinarizer
from sklearn.neighbors import NearestNeighbors
import scanpy as sc

import multiprocessing as mp
from tqdm import tqdm

from ..utils import convert_to_tensor, tensor_to_numpy
from ..utils import CustomDataset2

from typing import Literal, List, Tuple, Dict
from functools import partial


def codebook_generate(sure_model, n_samples):
    code_weights = convert_to_tensor(sure_model.codebook_weights, dtype=sure_model.dtype, device=sure_model.get_device())
    ns = dist.OneHotCategorical(probs=code_weights).sample([n_samples])

    codebook_loc, codebook_scale = sure_model.get_codebook()
    codebook_loc = convert_to_tensor(codebook_loc, dtype=sure_model.dtype, device=sure_model.get_device())
    codebook_scale = convert_to_tensor(codebook_scale, dtype=sure_model.dtype, device=sure_model.get_device())

    loc = torch.matmul(ns, codebook_loc)
    scale = torch.matmul(ns, codebook_scale)
    zs = dist.Normal(loc, scale).to_event(1).sample()
    return tensor_to_numpy(zs), tensor_to_numpy(ns)


def codebook_sample(sure_model, xs, n_samples, even_sample=False, filter=True):
    xs = convert_to_tensor(xs, dtype=sure_model.dtype, device=sure_model.get_device())
    assigns = sure_model.soft_assignments(xs)
    code_assigns = np.argmax(assigns, axis=1)
    
    if even_sample:        
        repeat = n_samples // assigns.shape[1]
        remainder = n_samples % assigns.shape[1]
        ns_id = np.repeat(np.arange(1, assigns.shape[1] + 1), repeat)
        # 补充剩余元素（将前 `remainder` 个数字各多重复1次）
        if remainder > 0:
            ns_id = np.concatenate([ns_id, np.arange(1, remainder + 1)])
        ns_id -= 1
        
        ns = LabelBinarizer().fit_transform(ns_id)
        ns = convert_to_tensor(ns, dtype=sure_model.dtype, device=sure_model.get_device())
    else:
        code_weights = codebook_weights(assigns)
        code_weights = convert_to_tensor(code_weights, dtype=sure_model.dtype, device=sure_model.get_device())
        ns = dist.OneHotCategorical(probs=code_weights).sample([n_samples])
        ns_id = np.argmax(tensor_to_numpy(ns), axis=1)

    codebook_loc, codebook_scale = sure_model.get_codebook()
    codebook_loc = convert_to_tensor(codebook_loc, dtype=sure_model.dtype, device=sure_model.get_device())
    codebook_scale = convert_to_tensor(codebook_scale, dtype=sure_model.dtype, device=sure_model.get_device())

    loc = torch.matmul(ns, codebook_loc)
    scale = torch.matmul(ns, codebook_scale)
    zs = dist.Normal(loc, scale).to_event(1).sample()

    xs_zs = sure_model.get_cell_coordinates(xs)
    #xs_zs = convert_to_tensor(xs_zs, dtype=sure_model.dtype, device=sure_model.get_device())
    #xs_dist = torch.cdist(zs, xs_zs)
    #idx = xs_dist.argmin(dim=1)
    
    #nbrs = NearestNeighbors(n_jobs=-1, n_neighbors=1)
    #nbrs.fit(tensor_to_numpy(xs_zs))
    #idx = nbrs.kneighbors(tensor_to_numpy(zs), return_distance=False)
    #idx_ = idx.flatten()
    #idx = [idx_[i] for i in np.arange(n_samples) if np.array_equal(code_assigns[idx_[i]], ns[i])]
    #df = pd.DataFrame({'idx':idx_,
    #                   'to':code_assigns[idx_],
    #                   'from':ns_id})
    #if filter:
    #    filtered_df = df[df['from'] != df['to']]
    #else:
    #    filtered_df = df
    #idx = filtered_df.loc[:,'idx'].values
    #ns_id = filtered_df.loc[:,'from'].values

    nbrs = NearestNeighbors(n_neighbors=50, n_jobs=-1)
    nbrs.fit(tensor_to_numpy(xs_zs))
        
    distances, ids = nbrs.kneighbors(tensor_to_numpy(zs), return_distance=True)
        
    idx,ns_list = [],[]
    with tqdm(total=n_samples, desc='Sketching', unit='sketch') as pbar:
        for i in np.arange(n_samples):
            distances_i = distances[i]
            weights_i = distance_to_softmax_weights(distances_i)
            cell_i_ = weighted_sample(ids[i], weights_i, sample_size=1, replace=False)
            
            df = pd.DataFrame({'idx':[cell_i_],
                               'to': [code_assigns[cell_i_]],
                               'from': [ns_id[i]]})
            if filter:
                filtered_df = df[df['from'] != df['to']]
            else:
                filtered_df = df
            cells_i = filtered_df.loc[:,'idx'].values
            ns_i = filtered_df.loc[:,'from'].unique()

            idx.extend(cells_i)
            ns_list.extend(ns_i)

            pbar.update(1)
            
    return tensor_to_numpy(xs[idx].squeece()), tensor_to_numpy(idx.squeeze()), ns_list


def codebook_sketch(sure_model, xs, n_samples, even_sample=False):
    return codebook_sample(sure_model, xs, n_samples, even_sample)

def codebook_bootstrap_sketch(sure_model, xs, n_samples, n_neighbors=8, 
                              aggregate_fun: Literal['mean','sum'] = 'mean', 
                              even_sample=False, replace=True, filter=True):
    xs = convert_to_tensor(xs, dtype=sure_model.dtype, device=sure_model.get_device())
    xs_zs = sure_model.get_cell_coordinates(xs)
    xs_zs = tensor_to_numpy(xs_zs)

    # generate samples that follow the metacell distribution of the given data
    assigns = sure_model.soft_assignments(xs)
    code_assigns = np.argmax(assigns,axis=1)
    if even_sample:
        repeat = n_samples // assigns.shape[1]
        remainder = n_samples % assigns.shape[1]
        ns_id = np.repeat(np.arange(1, assigns.shape[1] + 1), repeat)
        # 补充剩余元素（将前 `remainder` 个数字各多重复1次）
        if remainder > 0:
            ns_id = np.concatenate([ns_id, np.arange(1, remainder + 1)])
        ns_id -= 1
        
        ns = LabelBinarizer().fit_transform(ns_id)
        ns = convert_to_tensor(ns, dtype=sure_model.dtype, device=sure_model.get_device())
    else:
        code_weights = codebook_weights(assigns)
        code_weights = convert_to_tensor(code_weights, dtype=sure_model.dtype, device=sure_model.get_device())
        ns = dist.OneHotCategorical(probs=code_weights).sample([n_samples])
        ns_id = np.argmax(tensor_to_numpy(ns), axis=1)

    codebook_loc, codebook_scale = sure_model.get_codebook()
    codebook_loc = convert_to_tensor(codebook_loc, dtype=sure_model.dtype, device=sure_model.get_device())
    codebook_scale = convert_to_tensor(codebook_scale, dtype=sure_model.dtype, device=sure_model.get_device())

    loc = torch.matmul(ns, codebook_loc)
    scale = torch.matmul(ns, codebook_scale)
    zs = dist.Normal(loc, scale).to_event(1).sample()
    zs = tensor_to_numpy(zs)

    # find the neighbors of sample data in the real data space
    nbrs = NearestNeighbors(n_neighbors=50, n_jobs=-1)
    nbrs.fit(xs_zs)
        
    xs_list = []
    ns_list = []
    distances, ids = nbrs.kneighbors(zs, return_distance=True)
    #dist_pdf = gaussian_kde(distances.flatten())

    xs = tensor_to_numpy(xs)    
    sketch_cells = dict()
    with tqdm(total=n_samples, desc='Sketching', unit='sketch') as pbar:
        for i in np.arange(n_samples):
            #cells_i_ = ids[i, dist_pdf(distances[i]) > pval]
            #cells_i = [c for c in cells_i_ if np.array_equal(code_assigns[c],ns[i])]
            distances_i = distances[i]
            weights_i = distance_to_softmax_weights(distances_i)
            cells_i_ = weighted_sample(ids[i], weights_i, sample_size=n_neighbors, replace=replace)
            
            df = pd.DataFrame({'idx':cells_i_,
                               'to': code_assigns[cells_i_],
                               'from': [ns_id[i]] * len(cells_i_)})
            if filter:
                filtered_df = df[df['from'] != df['to']]
            else:
                filtered_df = df
            cells_i = filtered_df.loc[:,'idx'].values
            ns_i = filtered_df.loc[:,'from'].unique()

            if len(cells_i)>0:
                xs_i = xs[cells_i]
                if aggregate_fun == 'mean':
                    xs_i = np.mean(xs_i, axis=0, keepdims=True)
                elif aggregate_fun == 'median':
                    xs_i = np.median(xs_i, axis=0, keepdims=True)
                elif aggregate_fun == 'sum':
                    xs_i = np.sum(xs_i, axis=0, keepdims=True)

                xs_list.append(xs_i)
                ns_list.extend(ns_i)
                sketch_cells[i] = cells_i 

            pbar.update(1)

    return np.vstack(xs_list),sketch_cells,ns_list


def process_chunk(chunk_indices: np.ndarray,
                 zs: np.ndarray,
                 xs_zs: np.ndarray,
                 xs: np.ndarray,
                 code_assigns: np.ndarray,
                 ns_id: np.ndarray,
                 n_neighbors: int,
                 replace: bool,
                 filter: bool,
                 aggregate_fun: str) -> Tuple[List[np.ndarray], Dict[int, np.ndarray], List[int]]:
    """
    处理一个chunk的样本
    
    参数:
        chunk_indices: 当前chunk包含的样本索引
        其他参数与主函数相同
        
    返回:
        (xs_chunk, sketch_cells_chunk, ns_list_chunk)
    """
    xs_chunk = []
    sketch_cells_chunk = {}
    ns_list_chunk = []
    
    # 每个chunk创建自己的NearestNeighbors实例，避免多进程冲突
    nbrs = NearestNeighbors(n_neighbors=50, n_jobs=1)  # 单线程模式
    nbrs.fit(xs_zs)
    
    for i in chunk_indices:
        distances, ids = nbrs.kneighbors(zs[i:i+1], return_distance=True)
        distances_i = distances[0]
        weights_i = distance_to_softmax_weights(distances_i)
        cells_i_ = weighted_sample(ids[0], weights_i, sample_size=n_neighbors, replace=replace)
        
        df = pd.DataFrame({
            'idx': cells_i_,
            'to': code_assigns[cells_i_],
            'from': [ns_id[i]] * len(cells_i_)
        })
        
        if filter:
            filtered_df = df[df['from'] != df['to']]
        else:
            filtered_df = df
            
        cells_i = filtered_df.loc[:, 'idx'].values
        ns_i = filtered_df.loc[:, 'from'].unique()
        
        if len(cells_i) > 0:
            xs_i = xs[cells_i]
            
            if aggregate_fun == 'mean':
                xs_i = np.mean(xs_i, axis=0, keepdims=True)
            elif aggregate_fun == 'median':
                xs_i = np.median(xs_i, axis=0, keepdims=True)
            elif aggregate_fun == 'sum':
                xs_i = np.sum(xs_i, axis=0, keepdims=True)
                
            xs_chunk.append(xs_i)
            sketch_cells_chunk[i] = cells_i
            ns_list_chunk.extend(ns_i)
    
    return xs_chunk, sketch_cells_chunk, ns_list_chunk

def codebook_bootstrap_sketch_parallel(
    sure_model, 
    xs, 
    n_samples, 
    n_neighbors=8, 
    aggregate_fun: Literal['mean','sum'] = 'mean', 
    even_sample=False, 
    replace=True, 
    filter=True,
    n_processes: int = None,
    chunk_size: int = 100
) -> Tuple[np.ndarray, Dict[int, np.ndarray], List[int]]:
    """
    基于chunk的并行版本
    
    新增参数:
        n_processes: 并行进程数，None表示使用所有CPU核心
        chunk_size: 每个chunk包含的样本数
    """
    # 转换输入数据 (与原始版本相同)
    xs = convert_to_tensor(xs, dtype=sure_model.dtype, device=sure_model.get_device())
    xs_zs = sure_model.get_cell_coordinates(xs)
    xs_zs = tensor_to_numpy(xs_zs)

    # 生成样本 (与原始版本相同)
    assigns = sure_model.soft_assignments(xs)
    code_assigns = np.argmax(assigns, axis=1)
    
    if even_sample:
        repeat = n_samples // assigns.shape[1]
        remainder = n_samples % assigns.shape[1]
        ns_id = np.repeat(np.arange(1, assigns.shape[1] + 1), repeat)
        if remainder > 0:
            ns_id = np.concatenate([ns_id, np.arange(1, remainder + 1)])
        ns_id -= 1
        ns = LabelBinarizer().fit_transform(ns_id)
        ns = convert_to_tensor(ns, dtype=sure_model.dtype, device=sure_model.get_device())
    else:
        code_weights = codebook_weights(assigns)
        code_weights = convert_to_tensor(code_weights, dtype=sure_model.dtype, device=sure_model.get_device())
        ns = dist.OneHotCategorical(probs=code_weights).sample([n_samples])
        ns_id = np.argmax(tensor_to_numpy(ns), axis=1)

    # 获取codebook (与原始版本相同)
    codebook_loc, codebook_scale = sure_model.get_codebook()
    codebook_loc = convert_to_tensor(codebook_loc, dtype=sure_model.dtype, device=sure_model.get_device())
    codebook_scale = convert_to_tensor(codebook_scale, dtype=sure_model.dtype, device=sure_model.get_device())

    # 生成zs (与原始版本相同)
    loc = torch.matmul(ns, codebook_loc)
    scale = torch.matmul(ns, codebook_scale)
    zs = dist.Normal(loc, scale).to_event(1).sample()
    zs = tensor_to_numpy(zs)

    # 转换为numpy数组 (与原始版本相同)
    xs = tensor_to_numpy(xs)

    # 准备结果容器
    xs_list = []
    sketch_cells = {}
    ns_list = []

    # 分割样本为chunks
    chunks = [np.arange(i, min(i + chunk_size, n_samples)) 
              for i in range(0, n_samples, chunk_size)]
    
    # 创建进程池
    with mp.Pool(processes=n_processes) as pool:
        # 使用partial固定参数
        worker = partial(
            process_chunk,
            zs=zs,
            xs_zs=xs_zs,
            xs=xs,
            code_assigns=code_assigns,
            ns_id=ns_id,
            n_neighbors=n_neighbors,
            replace=replace,
            filter=filter,
            aggregate_fun=aggregate_fun
        )
        
        # 使用tqdm显示进度
        results = list(tqdm(
            pool.imap(worker, chunks),
            total=len(chunks),
            desc='Processing chunks',
            unit='chunk'
        ))
    
    # 合并结果
    for xs_chunk, sketch_cells_chunk, ns_list_chunk in results:
        xs_list.extend(xs_chunk)
        sketch_cells.update(sketch_cells_chunk)
        ns_list.extend(ns_list_chunk)

    return (
        np.vstack(xs_list) if xs_list else np.empty((0, xs.shape[1])),
        sketch_cells,
        ns_list
    )


def codebook_summarize_(assigns, xs):
    assigns = convert_to_tensor(assigns)
    xs = convert_to_tensor(xs)
    results = torch.matmul(assigns.T, xs)
    results = results / torch.sum(assigns.T, dim=1, keepdim=True)
    return tensor_to_numpy(results)


def codebook_summarize(assigns, xs, batch_size=1024):
    assigns = convert_to_tensor(assigns)
    xs = convert_to_tensor(xs)

    dataset = CustomDataset2(assigns, xs)
    dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)

    R = None
    W = None
    with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
        for A_batch, X_batch, _ in dataloader:
            r = torch.matmul(A_batch.T, X_batch)
            w = torch.sum(A_batch.T, dim=1, keepdim=True)
            if R is None:
                R = r 
                W = w 
            else:
                R += r 
                W += w 
            pbar.update(1)

    results = R / W
    return tensor_to_numpy(results)

def codebook_aggregate(assigns, xs, batch_size=1024):
    assigns = convert_to_tensor(assigns)
    xs = convert_to_tensor(xs)

    dataset = CustomDataset2(assigns, xs)
    dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)

    R = None
    with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
        for A_batch, X_batch, _ in dataloader:
            r = torch.matmul(A_batch.T, X_batch)
            if R is None:
                R = r 
            else:
                R += r 
            pbar.update(1)

    results = R
    return tensor_to_numpy(results)


def codebook_weights(assigns):
    assigns = convert_to_tensor(assigns)
    results = torch.sum(assigns, dim=0)
    results = results / torch.sum(results)
    return tensor_to_numpy(results)


def distance_to_softmax_weights(distances):
    """使用softmax将距离列表转换为概率权重
    
    参数:
        distances: 距离列表，距离越小权重应该越大
        
    返回:
        概率权重数组，和为1
    """
    distances = np.array(distances)
    # 取负数使得距离越小值越大
    negative_distances = -distances
    # 计算softmax
    exp_dist = np.exp(negative_distances - np.max(negative_distances))  # 数值稳定性处理
    softmax = exp_dist / np.sum(exp_dist)
    return softmax

def weighted_sample(items, weights, sample_size=1, replace=True):
    """根据权重进行采样
    
    参数:
        items: 待采样的列表
        weights: 对应的概率权重列表
        sample_size: 采样数量
        replace: 是否允许重复采样
        
    返回:
        采样结果列表
    """
    return np.random.choice(
        a=items,
        size=sample_size,
        p=weights,
        replace=replace
    )
    
def split_evenly(n, m):
    """使用NumPy将n分成m个尽可能平均的数"""
    arr = np.full(m, n // m)
    arr[:n % m] += 1
    return arr.tolist()


# Importing specific functions from modules
from .codebook import codebook_summarize_, codebook_summarize, codebook_aggregate, \
        codebook_generate, codebook_weights, codebook_sample, codebook_sketch, \
        codebook_bootstrap_sketch


import heapq

class PriorityQueue:
    def __init__(self):
        self.elements = []
        self.entry_finder = {}  # mapping of tasks to entries
        self.REMOVED = '<removed-task>'  # placeholder for a removed task
        self.counter = 0  # unique sequence count

    def is_empty(self):
        return not self.elements

    def put(self, item, priority):
        if item in self.entry_finder:
            self.remove(item)  # Remove the existing entry if it exists
        entry = [priority, self.counter, item]  # New entry with priority, counter, and item
        self.entry_finder[item] = entry
        heapq.heappush(self.elements, entry)
        self.counter += 1

    def remove(self, item):
        # Mark an existing task as REMOVED
        entry = self.entry_finder.pop(item)
        entry[-1] = self.REMOVED  # Mark it as removed

    def get(self):
        while self.elements:
            priority, count, item = heapq.heappop(self.elements)
            if item is not self.REMOVED:
                del self.entry_finder[item]  # Remove from the entry finder
                return item
        raise KeyError('pop from an empty priority queue')

    def peek(self):
        while self.elements:
            priority, count, item = self.elements[0]
            if item is not self.REMOVED:
                return item
            heapq.heappop(self.elements)  # Remove stale entry
        return None

    def update(self, item, priority):
        self.put(item, priority)  # Use put to add or update the item

    def __iter__(self):
        # Iterate over the priority queue without popping items
        # This will create a shallow copy of the current elements
        temp_elements = [(priority, count, item) for priority, count, item in self.elements if item is not self.REMOVED]
        for priority, count, item in sorted(temp_elements):
            yield item



import scanpy as sc
import pandas as pd
import numpy as np
import scipy as sp
from scipy.sparse import csr_matrix
from sklearn.preprocessing import OneHotEncoder, LabelBinarizer

import os 
import shutil
import tempfile
import subprocess 
import pkg_resources
import datatable as dt

from ..SURE import SURE 
from ..codebook import codebook_generate, codebook_summarize, codebook_aggregate, codebook_sketch, codebook_bootstrap_sketch
from ..utils import convert_to_tensor, tensor_to_numpy
from ..utils import pretty_print
from ..utils import find_partitions_greedy

from concurrent.futures import ThreadPoolExecutor, as_completed
from tqdm import tqdm


def assembly(adata_list_, batch_key, preprocessing=True, hvgs=None,
             n_top_genes=5000, hvg_method='cell_ranger', layer='counts', cuda_id=0, use_jax=False,
             codebook_size=800, codebook_size_per_adata=500, min_gamma=20, learning_rate=0.0001,
             batch_size=200, batch_size_per_adata=100, n_epochs=200, latent_dist='studentt',
             use_dirichlet=False, use_dirichlet_per_adata=False,
             zero_inflation=False, zero_inflation_per_adata=False,
             likelihood='multinomial', likelihood_per_adata='multinomial', 
             #n_samples_per_adata=10000, total_samples=1000000, summarize=False,
             sketch_func='sum', n_bootstrap_neighbors=8,
             #sketching=True, even_sketch=True, bootstrap_sketch=False, n_sketch_neighbors=10,
             n_workers=1, mute=True):
    adata_list = [ad.copy() for ad in adata_list_]
    n_adatas = len(adata_list)
    
    jit=''
    if use_jax:
        jit='--jit'

    # get common hvgs
    if hvgs is None:
        preprocessing = False if (hvg_method=='seurat_v3') else preprocessing
        
        # preprocess
        if preprocessing:
            with tqdm(total=n_adatas, desc='Preprocessing', unit='adata') as pbar:
                for i in np.arange(n_adatas):
                    #print(f'Adata {i+1} / {n_adatas}: Preprocessing')
                    adata_list[i] = preprocess(adata_list[i], layer)
                    pbar.update(1)

        #for i in np.arange(n_adatas):
        #    print(f'Adata {i+1} / {n_adatas}: Find {n_top_genes} HVGs')
        #    hvgs_ = highly_variable_genes(adata_list[i], n_top_genes, hvg_method)
        #    if hvgs is None:
        #        hvgs = set(hvgs_)
        #    else:
        #        hvgs = hvgs & set(hvgs_)
        #hvgs = list(hvgs)
        
        # test
        hvgs = highly_variable_genes_for_adatas(adata_list, n_top_genes, hvg_method)
        print(f'{len(hvgs)} common HVGs are found')
        
    adata_size_list = []
    for i in np.arange(n_adatas):
        adata_size_list.append(adata_list[i].shape[0])

    for i in np.arange(n_adatas):
        mask = [x in adata_list[i].var_names.tolist() for x in hvgs]
        if all(mask):
            adata_list[i] = adata_list[i][:,hvgs]    
        else:
            adata_i_X = get_subdata(adata_list[i], hvgs, 'X')
            adata_i_counts = get_subdata(adata_list[i], hvgs, 'counts')
            adata_i_obs = adata_list[i].obs
            adata_i_new = sc.AnnData(adata_i_X, obs=adata_i_obs)
            adata_i_new.var_names = hvgs
            adata_i_new.layers['counts'] = adata_i_counts
            adata_list[i] = adata_i_new

    models_list = []
    models_file_list = []
    model = None
    # process
    #with tempfile.TemporaryDirectory() as temp_dir:
    try:
        if not os.path.exists('./SURE_temp_dir'):
            os.mkdir('./SURE_temp_dir')
        else:
            pass
        temp_dir = tempfile.mkdtemp(dir='./SURE_temp_dir')
        
        if latent_dist == 'lapacian':
            latent_dist_param='--z-dist laplacian'
        elif latent_dist == 'studentt':
            latent_dist_param='--z-dist studentt'
        elif latent_dist == 'cauchy':
            latent_dist_param='--z-dist cauchy'
        else:
            latent_dist_param=''

        dirichlet = '-dirichlet' if use_dirichlet else ''
        dirichlet_per_adata = '-dirichlet' if use_dirichlet_per_adata else ''

        zi = '-zi exact' if zero_inflation else ''
        zi_per_adata = '-zi exact' if zero_inflation_per_adata else ''

        def run_sure(ad, i):
            X = get_data(ad, layer=layer)
            if batch_key is not None:
                U = batch_encoding(ad, batch_key=batch_key)
            else:
                U = pd.DataFrame(np.zeros((X.shape[0])), columns=['batch'])

            temp_count_file = os.path.join(temp_dir, f'temp_counts_{i}.txt.gz')
            temp_uwv_file = os.path.join(temp_dir, f'temp_uwv_{i}.txt.gz')
            temp_model_file = os.path.join(temp_dir, f'temp_{i}.pth')
            temp_log_file = os.path.join(temp_dir, f'temp_log_{i}.txt.gz')

            dt.Frame(X.round()).to_csv(temp_count_file)
            dt.Frame(U).to_csv(temp_uwv_file)
            
            codebook_size_per_adata_ = np.int32(X.shape[0] / min_gamma)
            codebook_size_per_adata_ = codebook_size_per_adata_ if codebook_size_per_adata > codebook_size_per_adata_ else codebook_size_per_adata
            
            mute_cmd = ''
            if mute:
                mute_cmd = f'2>&1 | gzip > {temp_log_file}'
            
            cmd = f'CUDA_VISIBLE_DEVICES={cuda_id}  SURE --data-file "{temp_count_file}" \
                        --undesired-factor-file "{temp_uwv_file}" \
                        --seed 0 \
                        --cuda {jit} \
                        -lr {learning_rate} \
                        -n {n_epochs} \
                        -bs {batch_size_per_adata} \
                        -cs {codebook_size_per_adata_} \
                        -likeli {likelihood_per_adata} {latent_dist_param} {dirichlet_per_adata} {zi_per_adata} \
                        --save-model "{temp_model_file}" {mute_cmd}'
            subprocess.call(f'{cmd}', shell=True)
            return {i:temp_model_file}
            
        with ThreadPoolExecutor(max_workers=n_workers) as executor:
            #models_file_list = list(tqdm(executor.map(run_sure, adata_list, np.arange(len(adata_list))), total=len(adata_list), desc="Metacell calling"))
            futures = [executor.submit(run_sure, ad, i) for i,ad in enumerate(adata_list)]
            results_ = {}
            for f in tqdm(as_completed(futures), total=len(adata_list), unit='adata', desc="Metacell calling"):
                results_.update(f.result())
            models_file_list = [results_[k] for k in sorted(results_)]
            
        # get the distribution structure for each adata
        for i in np.arange(n_adatas):
            '''
            X = get_data(adata_list[i], layer=layer)
            if batch_key is not None:
                U = batch_encoding(adata_list[i], batch_key=batch_key)
            else:
                U = pd.DataFrame(np.zeros((X.shape[0])), columns=['batch'])

            temp_count_file = os.path.join(temp_dir, f'temp_counts_{i}.txt.gz')
            temp_uwv_file = os.path.join(temp_dir, f'temp_uwv_{i}.txt.gz')
            temp_model_file = os.path.join(temp_dir, f'temp_{i}.pth')

            dt.Frame(X.round()).to_csv(temp_count_file)
            dt.Frame(U).to_csv(temp_uwv_file)
            
            codebook_size_per_adata_ = np.int32(X.shape[0] / min_gamma)
            codebook_size_per_adata_ = codebook_size_per_adata_ if codebook_size_per_adata > codebook_size_per_adata_ else codebook_size_per_adata
            
            print(f'Adata {i+1} / {n_adatas}: Compute distribution-preserved sketching with {codebook_size_per_adata_} metacells from {X.shape[0]} cells')

            cmd = f'CUDA_VISIBLE_DEVICES={cuda_id}  SURE --data-file "{temp_count_file}" \
                        --undesired-factor-file "{temp_uwv_file}" \
                        --seed 0 \
                        --cuda {jit} \
                        -lr {learning_rate} \
                        -n {n_epochs} \
                        -bs {batch_size_per_adata} \
                        -cs {codebook_size_per_adata_} \
                        -likeli {likelihood_per_adata} {latent_dist_param} {dirichlet_per_adata} {zi_per_adata} \
                        --save-model "{temp_model_file}" '
            pretty_print(cmd)
            subprocess.call(f'{cmd}', shell=True)
            model_i = SURE.load_model(temp_model_file)
            '''
            model_i = SURE.load_model(models_file_list[i])
            models_list.append(model_i)

        # generate samples from the learned distributions for assembly
        if n_adatas > 1:
            '''
            n_samples = n_samples_per_adata * n_adatas
            n_samples = n_samples if n_samples<total_samples else total_samples
            adata_freq = adata_size_list / np.sum(adata_size_list)
            #n_samples_list = generate_equal_list(n_samples, n_adatas)
            n_samples_list = n_samples * adata_freq
            n_samples_list = [np.int32(x) for x in n_samples_list]
            
            for i in np.arange(n_adatas):
                n_samples_list[i] = n_samples_list[i] if n_samples_list[i] < adata_size_list[i] else adata_size_list[i]
                n_samples_list[i] = n_samples_list[i] if n_samples_list[i] > min_gamma * models_list[i].code_size else min_gamma * models_list[i].code_size
            '''

            adatas_to_assembly=[]
            with tqdm(total=n_adatas, desc='Sampling', unit='adata') as pbar:
                for i in np.arange(n_adatas):
                #    print(f'Generate {n_samples_list[i]} samples from sketched atlas {i+1} / {n_adatas} ')
            
                    model_i = models_list[i]
                    if sketch_func=='sum':
                        xs_i_ = get_data(adata_list[i], layer=layer).values
                        ns = models_list[i].hard_assignments(xs_i_)
                        xs_i = codebook_aggregate(ns, xs_i_)
                    elif sketch_func=='mean':
                        xs_i_ = get_data(adata_list[i], layer=layer).values
                        ns = models_list[i].hard_assignments(xs_i_)
                        xs_i = codebook_summarize(ns, xs_i_)
                    elif sketch_func=='simulate':
                        zs_i,_ = codebook_generate(model_i, min_gamma * models_list[i].code_size)
                        xs_i = model_i.generate_count_data(zs_i)
                    elif sketch_func=='sample':
                        data_i = get_subdata(adata_list[i], hvgs=hvgs, layer=layer).values
                        xs_i,_,_ = codebook_sketch(model_i, data_i, min_gamma * models_list[i].code_size, even_sample=True)
                    elif sketch_func=='bootstrap_sample':
                        data_i = get_subdata(adata_list[i], hvgs=hvgs, layer=layer).values
                        xs_i,_,_ = codebook_bootstrap_sketch(model_i, data_i, min_gamma * models_list[i].code_size, n_bootstrap_neighbors, even_sample=True)
                    '''
                    if summarize:
                        xs_i_ = get_data(adata_list[i], layer=layer).values
                        ns = models_list[i].hard_assignments(xs_i_)
                        #xs_i = codebook_summarize(ns, xs_i_)
                        xs_i = codebook_aggregate(ns, xs_i_)
                    elif sketching:
                        data_i = get_subdata(adata_list[i], hvgs=hvgs, layer=layer).values
                        if not bootstrap_sketch:
                            xs_i,_,_ = codebook_sketch(model_i, data_i, n_samples_list[i], even_sample=even_sketch)
                        else:
                            xs_i,_,_ = codebook_bootstrap_sketch(model_i, data_i, n_samples_list[i], n_sketch_neighbors, even_sample=even_sketch)
                    else:
                        zs_i,_ = codebook_generate(model_i, n_samples_list[i])
                        xs_i = model_i.generate_count_data(zs_i)
                    '''

                    adata_i = sc.AnnData(xs_i)
                    adata_i.obs['adata_id'] = i

                    adatas_to_assembly.append(adata_i)
                    pbar.update(1)

            # assembly
            adata_to_assembly = sc.concat(adatas_to_assembly)
            temp_count_file = os.path.join(temp_dir, f'temp_counts.txt.gz')
            temp_uwv_file = os.path.join(temp_dir, f'temp_uwv.txt.gz')
            temp_model_file = os.path.join(temp_dir, f'temp_model.pth')

            X = get_data(adata_to_assembly, layer='X')
            U = batch_encoding(adata_to_assembly, batch_key='adata_id')
            dt.Frame(X.round()).to_csv(temp_count_file)
            dt.Frame(U).to_csv(temp_uwv_file)
            
            codebook_size_ = np.int32(X.shape[0] / min_gamma)
            codebook_size_ = codebook_size_ if codebook_size > codebook_size_ else codebook_size

            print(f'Create distribution-preserved atlas with {codebook_size_} metacells from {X.shape[0]} samples')
            cmd = f'CUDA_VISIBLE_DEVICES={cuda_id}  SURE --data-file "{temp_count_file}" \
                            --undesired-factor-file "{temp_uwv_file}" \
                            --seed 0 \
                            --cuda {jit}  \
                            -lr {learning_rate} \
                            -n {n_epochs} \
                            -bs {batch_size} \
                            -cs {codebook_size_} \
                            -likeli {likelihood} {latent_dist_param} {dirichlet} {zi} \
                            --save-model "{temp_model_file}" '
            pretty_print(cmd)
            subprocess.call(f'{cmd}', shell=True)
            model = SURE.load_model(temp_model_file)
        else:
            model = models_list[0]
    finally:
        shutil.rmtree('./SURE_temp_dir', ignore_errors=True)

    return model, models_list, hvgs

def preprocess(adata, layer='counts'):
    adata.X = get_data(adata, layer).values.copy()
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)
    return adata

def highly_variable_genes(adata, n_top_genes, hvg_method):
    sc.pp.highly_variable_genes(adata, n_top_genes=n_top_genes, flavor=hvg_method)
    hvgs = adata.var_names[adata.var.highly_variable]
    return hvgs

def highly_variable_genes_for_adatas__(adata_list, n_top_genes, hvg_method):
    n_adatas = len(adata_list)

    for i in np.arange(n_adatas):
        adata_list[i].obs['adata_id'] = i 

    adata = sc.concat(adata_list)
    sc.pp.highly_variable_genes(adata, n_top_genes=n_top_genes, flavor=hvg_method, batch_key='adata_id')
    hvgs = adata.var_names[adata.var.highly_variable].tolist()
    return hvgs

def highly_variable_genes_for_adatas(adata_list, n_top_genes, hvg_method):
    n_adatas = len(adata_list)

    dfs = []
    for i in np.arange(n_adatas):
        #print(f'Adata {i+1} / {n_adatas}: Find {n_top_genes} HVGs')
        adata = adata_list[i]
        sc.pp.highly_variable_genes(adata, n_top_genes=n_top_genes, flavor=hvg_method)
        hvgs_i = adata.var_names[adata.var.highly_variable].tolist()
        df_i = adata.var.loc[hvgs_i,:]
        df_i.reset_index(drop=False, inplace=True, names=["gene"])
        dfs.append(df_i)
        
    df = pd.concat(dfs, axis=0)
    df["highly_variable"] = df["highly_variable"].astype(int)
    df = df.groupby("gene", observed=True).agg(
        dict(
            means="mean",
            dispersions="mean",
            dispersions_norm="mean",
            highly_variable="sum",
        )
    )
    df["highly_variable_nbatches"] = df["highly_variable"]
    df["dispersions_norm"] = df["dispersions_norm"].fillna(0)

    df.sort_values(
            ["highly_variable_nbatches", "dispersions_norm"],
            ascending=False,
            na_position="last",
            inplace=True,
        )
    df["highly_variable"] = np.arange(df.shape[0]) < n_top_genes

    #df["highly_variable"] = (df["means"]>0.0125) & (df["means"]<3) & (df["dispersions_norm"]>0.5) & ((df["dispersions_norm"]<np.inf))

    df = df[df['highly_variable']]
    hvgs = df.index.tolist()
    #hvgs = hvgs[:n_top_genes]
    return hvgs

def highly_variable_genes_for_adatas_(adata_list, n_top_genes, hvg_method):
    n_adatas = len(adata_list)
    hvgs,hvgs_ = None,None

    for i in np.arange(n_adatas):
        print(f'Adata {i+1} / {n_adatas}: Find {n_top_genes} HVGs')
        adata = adata_list[i]
        sc.pp.highly_variable_genes(adata, n_top_genes=n_top_genes, flavor=hvg_method)
        if hvgs_:
            hvgs_ &= set(adata.var_names[adata.var.highly_variable].tolist())
        else:
            hvgs_ = set(adata.var_names[adata.var.highly_variable].tolist())
        
    if len(hvgs_) == n_top_genes:
        hvgs = list(hvgs_)
    else:
        hvgs = list(hvgs_)

        n_average = (n_top_genes - len(hvgs)) // n_adatas
        for i in np.arange(n_adatas):
            if i==n_adatas-1:
                n_average = n_top_genes - len(hvgs)

            adata = adata_list[i]
            hvgs_i = list(set(adata.var_names[adata.var.highly_variable].tolist()) - set(hvgs))
            df_i = adata.var.loc[hvgs_i,:].copy()
            #df_i.sort_values(by='highly_variable_rank', inplace=True)
            df_i.sort_values(by='dispersions_norm', ascending=False, inplace=True)
            hvgs_i = df_i.index.tolist()
            hvgs.extend(hvgs_i[:n_average])

    return hvgs

def batch_encoding_bk(adata, batch_key):
    sklearn_version = pkg_resources.get_distribution("scikit-learn").version
    if pkg_resources.parse_version(sklearn_version) < pkg_resources.parse_version("1.2"):
        enc = OneHotEncoder(sparse=False).fit(adata.obs[batch_key].to_numpy().reshape(-1,1))
    else:
        enc = OneHotEncoder(sparse_output=False).fit(adata.obs[batch_key].to_numpy().reshape(-1,1))
    return pd.DataFrame(enc.transform(adata.obs[batch_key].to_numpy().reshape(-1,1)), columns=enc.categories_[0])

def batch_encoding(adata, batch_key):
    enc = LabelBinarizer().fit(adata.obs[batch_key].to_numpy())
    return pd.DataFrame(enc.transform(adata.obs[batch_key].to_numpy()), columns=enc.classes_)
    
def get_data(adata, layer='counts'):
    if layer.lower()!='x':
        data = adata.layers[layer]
    else:
        data = adata.X

    if sp.sparse.issparse(data):
        data = data.toarray()

    data[np.isnan(data)] = 0

    return pd.DataFrame(data.astype('float32'), columns=adata.var_names) 

def get_subdata(adata, hvgs, layer='counts'):
    #mask = [hvg in X for hvg in hvgs]
    hvgs_df = pd.DataFrame({'hvgs':hvgs})
    mask = hvgs_df['hvgs'].isin(adata.var_names.tolist())
    if all(mask):
        X = get_data(adata[:,hvgs], layer)
        return X[hvgs]
    else:
        #X2 = np.zeros((X.shape[0], len(hvgs)))
        #X2 = pd.DataFrame(X2, columns=hvgs)

        #columns = [c for c in X.columns.tolist() if c in hvgs]
        #X2[columns] = X[columns].copy()

        # inspired by SCimilarity
        shell = sc.AnnData(
            X=csr_matrix((0, len(hvgs))),
            var=pd.DataFrame(index=hvgs),
        )
        if layer.lower() != 'x':
            shell.layers[layer] = shell.X.copy()
        shell = sc.concat(
            (shell, adata[:, adata.var.index.isin(shell.var.index)]), join="outer"
        )
        X2 = get_data(shell, layer)
        return X2[hvgs]
    
def get_uns(adata, key):
    data = None 

    if key in adata.uns:
        data = adata.uns[key]
        columns = adata.uns[f'{key}_columns']
        if sp.sparse.issparse(data):
            data = data.toarray()
        data = pd.DataFrame(data.astype('float32'), 
                            columns=columns) 
        
    return data
    
def split_by(adata, by:str, copy:bool=False):
    adata_list = []
    for id in adata.obs[by].unique():
        if copy:
            adata_list.append(adata[adata.obs[by].isin([id])].copy())
        else:
            adata_list.append(adata[adata.obs[by].isin([id])])
    return adata_list

def split_batch_by_bk(adata, by, batch_size=30000, copy=False):
    df = adata.obs[by].value_counts().reset_index()
    df.columns = [by,'Value']

    n = int(np.round(adata.shape[0] / batch_size))
    n = n if n > 0 else 1

    adata_list = []
    parts = find_partitions_greedy(df['Value'].tolist(), n)
    for _,by_ids in parts:
        ids = df.iloc[by_ids,:][by].tolist()
        if copy:
            adata_list.append(adata[adata.obs[by].isin(ids)].copy())
        else:
            adata_list.append(adata[adata.obs[by].isin(ids)])

    return adata_list

def split_batch_by(adata, 
                   by:str, 
                   batch_size: int = 30000, 
                   copy: bool = False):
    groups = adata.obs[by].unique()
    adata_list = []
    for grp in groups:
        adata_list.extend(split_batch(adata[adata.obs[by]==grp], batch_size=batch_size, copy=copy))

    return adata_list

def split_batch(adata, batch_size: int=30000, copy: bool=False):
    n = int(np.round(adata.shape[0] / batch_size))
    n = n if n > 0 else 1
    
    cells = adata.obs_names.tolist()
    chunks = np.array_split(cells, n)
    
    adata_list = []
    for chunk in chunks:
        chunk = list(chunk)
        
        if copy:
            adata_list.append(adata[chunk].copy())
        else:
            adata_list.append(adata[chunk])

    return adata_list


def generate_equal_list(total, n):
    base_value = total // n
    remainder = total - (base_value * n)
    result = [base_value] * (n - 1) + [base_value + remainder]
    return result


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='SUREv1',
    version='1.0.1',
    description='Succinct Representation of Single Cells',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Feng Zeng',
    author_email='zengfeng@xmu.edu.cn',
    packages=find_packages(),
    install_requires=['dill==0.3.8','scanpy','pytorch-ignite','datatable','scipy','numpy','scikit-learn','pandas','pyro-ppl',
                      'leidenalg','python-igraph','networkx','matplotlib','seaborn','fa2-modified'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
    url='https://github.com/ZengFLab/SURE',  # 项目的 GitHub 地址

    entry_points={
        'console_scripts': [
            'SURE=SURE.SURE:main'  # 允许用户通过命令行调用 main 函数
        ],
    },
)


import torch
from torch.utils.data import Dataset

import numpy as np
import textwrap

class CustomDataset(Dataset):
    def __init__(self, X):
        self.X = X

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        x = self.X[idx]
        return x, idx
    
class CustomDataset2(Dataset):
    def __init__(self, X, U):
        self.X = X
        self.U = U

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        x = self.X[idx]
        if self.U is None:
            u = x
        else:
            u = self.U[idx]
        return x, u, idx
    

class CustomDataset3(Dataset):
    def __init__(self, X, U, Y):
        self.X = X
        self.U = U
        self.Y = Y

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        x = self.X[idx]
        if self.U is None:
            u = x
        else:
            u = self.U[idx]
        if self.Y is None:
            y = x
        else:
            y = self.Y[idx]
        return x, u, y, idx
    
class CustomDataset4(Dataset):
    def __init__(self, X, Y, Z, U):
        self.X = X
        self.U = U
        self.Y = Y
        self.Z = Z

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        x = self.X[idx]
        if self.U is None:
            u = x
        else:
            u = self.U[idx]
        if self.Y is None:
            y = x
        else:
            y = self.Y[idx]
        if self.Z is None:
            z = x
        else:
            z = self.Z[idx]
        return x, y, z, u, idx

class CustomMultiOmicsDataset(Dataset):
    def __init__(self, X1, X2):
        self.X1 = X1
        self.X2 = X2

    def __len__(self):
        return len(self.X1)

    def __getitem__(self, idx):
        x1 = self.X1[idx]
        x2 = self.X2[idx]
        return x1, x2, idx
    
class CustomMultiOmicsDataset2(Dataset):
    def __init__(self, X1, X2, U):
        self.X1 = X1
        self.X2 = X2
        self.U = U

    def __len__(self):
        return len(self.X1)

    def __getitem__(self, idx):
        x1 = self.X1[idx]
        x2 = self.X2[idx]
        if self.U is None:
            u=x1
        else:
            u = self.U[idx]
        return x1, x2, u, idx
    
class CustomMultiOmicsDataset3(Dataset):
    def __init__(self, X1, X2, Y, U):
        self.X1 = X1
        self.X2 = X2
        self.U = U
        self.Y = Y

    def __len__(self):
        return len(self.X1)

    def __getitem__(self, idx):
        x1 = self.X1[idx]
        x2 = self.X2[idx]
        if self.U is None:
            u=x1
        else:
            u = self.U[idx]
        if self.Y is None:
            y=x1 
        else:
            y=self.Y[idx]
        return x1, x2, y, u, idx

class CustomMultiOmicsDataset4(Dataset):
    def __init__(self, X1, X2, Y, Z, U):
        self.X1 = X1
        self.X2 = X2
        self.U = U
        self.Y = Y
        self.Z = Z

    def __len__(self):
        return len(self.X1)

    def __getitem__(self, idx):
        x1 = self.X1[idx]
        x2 = self.X2[idx]
        if self.U is None:
            u=x1
        else:
            u = self.U[idx]
        if self.Y is None:
            y=x1 
        else:
            y = self.Y[idx]
        if self.Z is None:
            z=x1 
        else:
            z = self.Z[idx]
        return x1, x2, y, z, u, idx
    
def tensor_to_numpy(tensor):
    """
    Check if the tensor is on a CUDA device. If yes, detach it, move it to CPU,
    and convert to a NumPy array. If not, just detach and convert to NumPy.

    Args:
        tensor (torch.Tensor): The input tensor.

    Returns:
        np.ndarray: The resulting NumPy array.
    """
    # Check if the input is a tensor
    if not isinstance(tensor, torch.Tensor):
        if isinstance(tensor, np.ndarray):
            return tensor
        raise ValueError("Input must be a torch Tensor.")

    # Detach the tensor from the computation graph
    tensor = tensor.detach()
    
    # Check if the tensor is on CUDA
    if tensor.is_cuda:
        tensor = tensor.cpu()

    # Convert to NumPy
    numpy_array = tensor.numpy()
    return numpy_array

def move_to_device(data, device):
    """
    Checks if the input data is a tensor. If not, converts it to a tensor,
    checks if the tensor is on the specified device, and moves it if necessary.

    Args:
        data (any): The input data to check (can be a tensor, list, NumPy array, etc.).
        device (str or torch.device): The device to check against (e.g., 'cpu', 'cuda', 'cuda:0').

    Returns:
        torch.Tensor: The tensor on the specified device.
    """
    # Convert input data to tensor if it's not already a tensor
    if not isinstance(data, torch.Tensor):
        data = torch.tensor(data)

    # Check if the device is a string, and convert it to torch.device if necessary
    device = torch.device(device) if isinstance(device, str) else device

    # Move the tensor to the specified device if necessary
    if data.device != device:
        data = data.to(device)
    
    return data


def convert_to_tensor(input_array, dtype=torch.float32, device=None):
    """
    Check if the input array is a torch tensor and convert it to a tensor if it is not.
    If dtype is specified, convert the tensor to the specified dtype if necessary.
    
    Parameters:
    - input_array: The input array to check and convert.
    - dtype: The desired data type for the resulting tensor (optional).
    
    Returns:
    - A torch tensor.
    """
    # Check if the input is already a torch tensor
    if isinstance(input_array, torch.Tensor):
        #print("Input is already a torch tensor.")
        # If dtype is specified, check and convert if necessary
        if dtype is not None and input_array.dtype != dtype:
            #print(f"Changing tensor dtype from {input_array.dtype} to {dtype}.")
            input_array = input_array.to(dtype)
        if device:
            input_array = move_to_device(input_array, device)
        return input_array  # Return the tensor unchanged if dtype matches

    else:
        # Convert to torch tensor
        #print("Input is not a torch tensor. Converting to torch tensor.")
        tensor = torch.tensor(input_array, dtype=dtype)
        if device:
            tensor = move_to_device(tensor, device)
        return tensor
    

class Colors:
    RESET = "\033[0m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"

def pretty_print(long_text, width=120, color='green'):
    # Convert multiple spaces to a single space
    formatted_text = ' '.join(long_text.split())

    # Wrap the text to a specified width
    wrapped_text = textwrap.fill(formatted_text, width=width)

    # Define the indent for subsequent lines
    indent = '    '  # Four spaces for indentation

    # Split the wrapped text into lines
    lines = wrapped_text.split('\n')

    text_color = Colors.RESET
    if color.lower() == 'green':
        text_color = Colors.GREEN
    elif color.lower == 'yellow':
        text_color = Colors.YELLOW

    # Print the first line without indent
    print(text_color + lines[0] + Colors.RESET)

    # Print the subsequent lines with indent
    for line in lines[1:]:
        print(indent + text_color + line + Colors.RESET)

def find_partitions_greedy(numbers, num_groups):
    # Step 1: Calculate the target sum per group
    total_sum = sum(numbers)
    target_per_group = total_sum / num_groups

    # Initialize data structures
    groups = [[] for _ in range(num_groups)]  # Groups of numbers
    sums = [0] * num_groups  # Sums of each group
    indices = [[] for _ in range(num_groups)]  # Indices of numbers in original list

    # Step 2: Sort numbers and their indices based on value
    sorted_numbers_with_indices = sorted(enumerate(numbers), key=lambda x: -x[1])

    # Step 3: Distribute numbers to approach the target sum per group as close as possible
    for index, number in sorted_numbers_with_indices:
        # Find the group with the minimum sum
        min_group_index = sums.index(min(sums))
        groups[min_group_index].append(number)
        indices[min_group_index].append(index)
        sums[min_group_index] += number

    # Return the groups with their original indices
    return [(group, index_group) for group, index_group in zip(groups, indices)]